"""Matplotlib wrapper functions for visualization."""

from copy import copy
from typing import Any

import matplotlib as mpl
import numpy as np
from numpy import ndarray

from .._logging import logger
from .interpolation import Extent


def plot(*, x: ndarray, y: ndarray, ax: Any, **kwargs):
    """Plot particles.

    Parameters
    ----------
    x
        The x-coordinates for the particle plot.
    y
        The y-coordinates for the particle plot.
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.plot method.

    Returns
    -------
    lines
        A list of matplotlib Line2D objects.
    """
    _kwargs = copy(kwargs)
    linestyle = _kwargs.pop('linestyle', '')
    marker = _kwargs.pop('marker', '.')
    return ax.plot(x, y, linestyle=linestyle, marker=marker, **_kwargs)


def scatter(
    *,
    x: ndarray,
    y: ndarray,
    c: ndarray = None,
    s: ndarray = None,
    n_samples=10_000,
    random_seed=None,
    ax: Any,
    **kwargs,
):
    """Plot particles as scatter plot.

    Parameters
    ----------
    x
        The x-coordinates for the scatter plot.
    y
        The y-coordinates for the scatter plot.
    c
        The quantity to color the particles.
    s
        The quantity to set the particle size.
    n_samples
        The number of samples to take. Default is 10,000.
    random_seed
        The random seed for sampling. Default is None.
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.scatter method.

    Returns
    -------
    paths
        A matplotlib PathCollection object.
    """
    if c is None and s is None:
        raise ValueError('Should set size or color')
    if n_samples > 100_000:
        logger.warning('n_samples > 100,000: this may be slow')
    if random_seed is not None:
        np.random.seed(random_seed)

    rand = np.random.choice(len(x), n_samples)

    x = x[rand]
    y = y[rand]
    if c is not None:
        c = c[rand]
    if s is not None:
        s = s[rand]

    _kwargs = copy(kwargs)
    alpha = _kwargs.pop('alpha', 0.5)

    return ax.scatter(x, y, c=c, s=s, alpha=alpha, **_kwargs)


def imshow(*, interpolated_data: ndarray, extent: Extent, ax: Any, **kwargs):
    """Plot 1d interpolated data as an image.

    Parameters
    ----------
    interpolated_data
        The data interpolated to a pixel grid.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax).
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.imshow method.

    Returns
    -------
    image
        A matplotlib AxesImage object.
    """
    _kwargs = copy(kwargs)
    try:
        norm = _kwargs.pop('norm')
    except KeyError:
        norm = 'linear'
    if norm.lower() in ('linear', 'lin'):
        norm = mpl.colors.Normalize()
    elif norm.lower() in ('logarithic', 'logarithm', 'log', 'log10'):
        norm = mpl.colors.LogNorm()
    else:
        raise ValueError('Cannot determine normalization for colorbar')

    return ax.imshow(
        interpolated_data, origin='lower', extent=extent, norm=norm, **_kwargs
    )


def contour(*, interpolated_data: ndarray, extent: Extent, ax: Any, **kwargs):
    """Plot 1d interpolated data as a contour plot.

    Parameters
    ----------
    interpolated_data
        The data interpolated to a pixel grid.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax).
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.imshow method.

    Returns
    -------
    contour
        A matplotlib QuadContourSet object.
    """
    n_interp_x, n_interp_y = interpolated_data.shape
    X, Y = np.meshgrid(
        np.linspace(*extent[:2], n_interp_x), np.linspace(*extent[2:], n_interp_y),
    )

    return ax.contour(X, Y, interpolated_data, **kwargs)


def quiver(*, interpolated_data: ndarray, extent: Extent, ax: Any, **kwargs):
    """Plot 2d interpolated data as a quiver plot.

    Parameters
    ----------
    interpolated_data
        The data interpolated to a pixel grid.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax).
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.imshow method.

    Returns
    -------
    quiver
        A matplotlib Quiver object.
    """
    n_interp_x, n_interp_y = interpolated_data[0].shape
    X, Y = np.meshgrid(
        np.linspace(*extent[:2], n_interp_x), np.linspace(*extent[2:], n_interp_y)
    )
    U, V = interpolated_data[0], interpolated_data[1]

    _kwargs = copy(kwargs)
    number_of_arrows = _kwargs.pop('number_of_arrows', (25, 25))
    normalize_vectors = _kwargs.pop('normalize_vectors', False)

    n_x, n_y = number_of_arrows[0], number_of_arrows[1]
    stride_x = int(n_interp_x / n_x)
    stride_y = int(n_interp_y / n_y)
    X = X[::stride_y, ::stride_x]
    Y = Y[::stride_y, ::stride_x]
    U = U[::stride_y, ::stride_x]
    V = V[::stride_y, ::stride_x]
    if normalize_vectors:
        norm = np.hypot(U, V)
        U /= norm
        V /= norm

    return ax.quiver(X, Y, U, V, **_kwargs)


def streamplot(*, interpolated_data: ndarray, extent: Extent, ax: Any, **kwargs):
    """Plot 2d interpolated data as a stream plot.

    Parameters
    ----------
    interpolated_data
        The data interpolated to a pixel grid.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax).
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.imshow method.

    Returns
    -------
    streamplot
        A matplotlib StreamplotSet object.
    """
    n_interp_x, n_interp_y = interpolated_data[0].shape
    X, Y = np.meshgrid(
        np.linspace(*extent[:2], n_interp_x), np.linspace(*extent[2:], n_interp_y)
    )
    U, V = interpolated_data[0], interpolated_data[1]

    return ax.streamplot(X, Y, U, V, **kwargs)



"""SPH sums over neighbours."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple

import numba
import numpy as np
from numba.typed import List

from .._logging import logger
from .._units import Quantity
from ..utils.kernels import (
    kernel_cubic,
    kernel_gradient_cubic,
    kernel_gradient_quintic,
    kernel_gradient_wendland_c4,
    kernel_quintic,
    kernel_wendland_c4,
)

if TYPE_CHECKING:
    from ..snap import Snap


def derivative(
    snap: Snap,
    derivative: str,
    quantity: str,
    kernel: str = 'cubic',
    chunk_size: int = None,
    verbose: bool = False,
) -> Quantity:
    """Calculate derivatives of quantities.

    WARNING: This function is experimental.

    Calculate derivatives such as grad, div, curl on any particle
    quantity using the SPH kernel gradient.

    Parameters
    ----------
    snap
        The Snap object.
    derivative
        Options are 'grad', 'div', 'curl'.
    quantity
        The quantity to take the derivative of. Must be a string to pass
        to Snap object which returns a scalar.
    kernel
        Kernel to compute density. E.g. 'cubic', 'quintic', or
        'Wendland C4'.
    chunk_size : optional
        The size of chunks, in terms of particle number, for neighbour
        finding. If the chunk size is too large then the neighbour
        finding algorithm (scipy.spatial.cKDTree.query_ball_point) runs
        out of memory. Default is None.
    verbose : optional
        If True, print progress. Default is False.

    Returns
    -------
    Quantity
        The derivative of the quantity.
    """
    logger.warning('sph.derivative is experimental')
    logger.info(f'Calculating {derivative}... may take some time...')

    if derivative not in ('grad', 'div', 'curl'):
        raise ValueError('derivative must be in ("grad", "div", "curl")')

    density: Quantity = snap['density']
    quantity_array: Quantity = snap[quantity]

    compute_function = _compute_derivative
    compute_function_kwargs = {
        'density': density.magnitude,
        'quantity_array': quantity_array.magnitude,
    }

    if derivative == 'grad':
        if quantity_array.ndim > 1:
            raise ValueError('Quantity must be scalar for grad')
        result_shape = (len(snap), 3)
        compute_function_kwargs[
            'compute_derivative_over_neighbours'
        ] = _compute_grad_over_neighbours

    elif derivative == 'div':
        if quantity_array.ndim != 2:
            raise ValueError('Quantity must be vector for div')
        result_shape = (len(snap), 1)
        compute_function_kwargs[
            'compute_derivative_over_neighbours'
        ] = _compute_div_over_neighbours

    elif derivative == 'curl':
        if quantity_array.ndim != 2:
            raise ValueError('Quantity must be vector for curl')
        result_shape = (len(snap), 3)
        compute_function_kwargs[
            'compute_derivative_over_neighbours'
        ] = _compute_curl_over_neighbours

    compute_function_kwargs['result_axis_1_size'] = result_shape[1]

    result = summation(
        snap=snap,
        result_shape=result_shape,
        compute_function=compute_function,
        compute_function_kwargs=compute_function_kwargs,
        kernel=kernel,
        chunk_size=chunk_size,
        verbose=verbose,
    )

    logger.info(f'Calculating {derivative}... Done!')

    if derivative == 'div':
        return np.squeeze(result)
    return result


def summation(
    snap: Snap,
    result_shape: Tuple[int, ...],
    compute_function: Callable,
    compute_function_kwargs: Dict[str, Any],
    kernel: str = 'cubic',
    chunk_size: int = None,
    verbose: bool = False,
) -> Quantity:
    """Calculate SPH sums.

    WARNING: This function is experimental.

    Parameters
    ----------
    snap
        The Snap object.
    result_shape
        The shape, as a tuple, of the returned sum.
    compute_function
        The function that computes the sums.
    compute_function_kwargs
        The keyword arguments to the function that computes the sums.
    kernel
        Kernel to compute density. E.g. 'cubic', 'quintic', or
        'Wendland C4'.
    chunk_size : optional
        The size of chunks, in terms of particle number, for neighbour
        finding. If the chunk size is too large then the neighbour
        finding algorithm (scipy.spatial.cKDTree.query_ball_point) runs
        out of memory. Default is None.
    verbose : optional
        If True, print progress. Default is False.

    Returns
    -------
    Quantity
        The result of the SPH summation.
    """
    logger.warning('sph.summation is experimental')
    if kernel not in ('cubic', 'quintic', 'Wendland C4'):
        raise ValueError('Kernel must be in ("cubic", "quintic", "Wendland C4")')
    if kernel == 'cubic':
        kernel_function = kernel_cubic
        kernel_gradient_function = kernel_gradient_cubic
    elif kernel == 'quintic':
        kernel_function = kernel_quintic
        kernel_gradient_function = kernel_gradient_quintic
    elif kernel == 'Wendland C4':
        kernel_function = kernel_wendland_c4
        kernel_gradient_function = kernel_gradient_wendland_c4

    snap.set_kernel(kernel)

    position: Quantity = snap['position']
    smoothing_length: Quantity = snap['smoothing_length']
    mass: Quantity = snap['mass']

    _position = position.magnitude
    _smoothing_length = smoothing_length.magnitude
    _mass = mass.magnitude

    result = np.zeros(result_shape)

    for particle_type in snap.particle_type:
        if verbose:
            logger.info(f'Summing over {particle_type} particles...')

        ind = snap.particle_indices(particle_type)
        if isinstance(ind, list):
            _indices = ind
        else:
            _indices = [ind]

        for type_indices in _indices:
            if len(type_indices) == 0:
                continue
            if chunk_size is not None:
                n_chunks = int(len(type_indices) / chunk_size)
                array_chunks = np.array_split(type_indices, n_chunks)
            else:
                n_chunks = 1
                array_chunks = [type_indices]

            if verbose and n_chunks > 1:
                logger.info(f'Number of chunks: {n_chunks}')
                logger.info(f'Chunk size: {chunk_size}')

            for idx, indices in enumerate(array_chunks):
                if verbose:
                    if n_chunks > 1:
                        logger.info(f'Finding neighbours for chunk: {idx}...')
                    else:
                        logger.info('Finding neighbours...')
                _neighbours = snap[indices].neighbours(np.arange(len(indices)))
                neighbours = List()
                for neigh in _neighbours:
                    neighbours.append(np.array(neigh))

                if verbose:
                    if n_chunks > 1:
                        logger.info(f'Summing over neighbours for chunk: {idx}...')
                    else:
                        logger.info('Summing over neighbours...')

                result[indices] = compute_function(
                    indices=indices,
                    neighbours=neighbours,
                    type_indices=type_indices,
                    position=_position,
                    smoothing_length=_smoothing_length,
                    mass=_mass,
                    kernel_function=kernel_function,
                    kernel_gradient_function=kernel_gradient_function,
                    verbose=verbose,
                    **compute_function_kwargs,
                )

    return result


@numba.njit
def _compute_derivative(
    indices,
    neighbours,
    type_indices,
    position,
    smoothing_length,
    mass,
    kernel_function,
    kernel_gradient_function,
    density,
    quantity_array,
    compute_derivative_over_neighbours,
    result_axis_1_size,
    verbose,
):
    num_particles = len(neighbours)
    result = np.zeros((num_particles, result_axis_1_size))

    for idxi in range(num_particles):
        index = indices[idxi]

        percent = idxi / num_particles * 100
        if np.mod(percent, 10) == 0 and verbose:
            print(percent, '%')

        hi = smoothing_length[index]
        if hi < 0:
            continue

        __neigh = neighbours[idxi]
        _neigh = type_indices[__neigh]
        neigh = _neigh[(_neigh != index) & (smoothing_length[_neigh] > 0)]

        posi = position[index]
        quani = quantity_array[index]

        result[idxi] = compute_derivative_over_neighbours(
            posi=posi,
            hi=hi,
            quani=quani,
            quanj=quantity_array[neigh],
            posj=position[neigh],
            mj=mass[neigh],
            rhoj=density[neigh],
            kernel_gradient_function=kernel_gradient_function,
        )

    return result


@numba.njit
def _compute_grad_over_neighbours(
    posi, hi, quani, quanj, posj, mj, rhoj, kernel_gradient_function,
):
    result = np.zeros(3)

    for idxj in range(len(quanj)):
        dr = posi - posj[idxj]
        rij = np.sqrt(dr[0] ** 2 + dr[1] ** 2 + dr[2] ** 2)
        dr = dr / rij
        qi = rij / hi
        grad_kern = kernel_gradient_function(qi)
        result += (
            mj[idxj] / rhoj[idxj] * (quanj[idxj] - quani) * dr * grad_kern / hi ** 4
        )

    return result


@numba.njit
def _compute_div_over_neighbours(
    posi, hi, quani, quanj, posj, mj, rhoj, kernel_gradient_function,
):
    result = 0.0

    for idxj in range(len(quanj)):
        dq = quanj[idxj] - quani
        dr = posi - posj[idxj]
        rij = np.sqrt(dr[0] ** 2 + dr[1] ** 2 + dr[2] ** 2)
        dr = dr / rij
        qi = rij / hi
        grad_kern = kernel_gradient_function(qi)
        dot = dq[0] * dr[0] + dq[1] * dr[1] + dq[2] * dr[2]
        result += mj[idxj] / rhoj[idxj] * dot * grad_kern / hi ** 4

    return result


@numba.njit
def _compute_curl_over_neighbours(
    posi, hi, quani, quanj, posj, mj, rhoj, kernel_gradient_function,
):
    result = np.zeros(3)
    cross = np.zeros(3)

    for idxj in range(len(quanj)):
        dq = quanj[idxj] - quani
        dr = posi - posj[idxj]
        rij = np.sqrt(dr[0] ** 2 + dr[1] ** 2 + dr[2] ** 2)
        dr = dr / rij
        qi = rij / hi
        grad_kern = kernel_gradient_function(qi)
        cross[0] = dq[1] * dr[2] - dq[2] * dr[1]
        cross[1] = dq[2] * dr[0] - dq[0] * dr[2]
        cross[2] = dq[0] * dr[1] - dq[1] * dr[0]
        result += mj[idxj] / rhoj[idxj] * cross * grad_kern / hi ** 4

    return result



"""Read Phantom evolution files."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from pandas import DataFrame

from .._config import load_config
from .._units import _convert_dim_string, _get_code_unit, generate_array_code_units

if TYPE_CHECKING:
    from .simulation import Simulation


def load_data_from_file(
    filenames: Union[str, Path, Tuple[str], Tuple[Path], List[str], List[Path]],
    config: Union[str, Path] = None,
):
    """Load data from Phantom .ev files.

    Parameters
    ----------
    filenames
        The filename or filenames (as a list).
    config : optional
        The path to a Plonk config.toml file.

    Returns
    -------
    DataFrame
    """
    if isinstance(filenames, (str, Path)):
        _filenames = [filenames]
    elif isinstance(filenames, (list, tuple)):
        _filenames = list(filenames)
    else:
        raise ValueError('filenames is not a known type')

    conf = load_config(filename=config)
    name_map = conf['phantom']['time_series']['namemap']

    _file_paths = list()
    for filename in _filenames:
        path = Path(filename)
        _file_paths.append(path.resolve())
    file_paths = tuple(_file_paths)

    _check_file_consistency(filenames=file_paths, name_map=name_map)
    columns = _get_columns(filename=file_paths[0], name_map=name_map)
    dataframe = _get_data(columns=columns, file_paths=file_paths)

    return dataframe


def evolution_units(sim: Simulation, config: Union[str, Path] = None):
    """Get units of Phantom .ev files from Simulation object.

    Parameters
    ----------
    sim
        The Simulation object.
    config : optional
        The path to a Plonk config.toml file.

    Returns
    -------
    Dict
    """
    conf = load_config(filename=config)
    arrays = conf['phantom']['time_series']['dimensions']
    dim = dict()
    for key, val in arrays.items():
        dim[key] = _convert_dim_string(val)
    _units = dict()
    for arr, unit in dim.items():
        _units[arr] = _get_code_unit(unit, sim.code_units)
    return _units


def _get_data(columns: Tuple[str, ...], file_paths: Tuple[Path, ...]) -> DataFrame:

    times = list()
    for filename in file_paths:
        times.append(np.loadtxt(filename, usecols=0))

    _skiprows = [0]
    if len(times) > 1:
        for t1, t2 in zip(times, times[1:]):
            _skiprows.append(np.where(t2 < t1[-1])[0][-1] + 2)

    df = pd.concat(
        (
            pd.read_csv(
                f,
                names=columns,
                skiprows=skiprows,
                skipinitialspace=True,
                delim_whitespace=True,
                comment='#',
            )
            for f, skiprows in zip(file_paths, _skiprows)
        )
    )

    df.reset_index(inplace=True, drop=True)
    return df


def _get_columns(filename: Path, name_map: Dict[str, str]) -> Tuple[str, ...]:

    with open(filename) as f:
        column_line = f.readline().strip('\n')

    _column_line = [item.strip('] ')[2:].strip(' ') for item in column_line.split('[')]
    columns = _column_line[1:]

    return tuple([name_map[col] if col in name_map else col for col in columns])


def _check_file_consistency(
    filenames: Tuple[Path, ...], name_map: Dict[str, str]
) -> None:

    columns = _get_columns(filenames[0], name_map)
    for filename in filenames:
        columns_previous = columns
        columns = _get_columns(filename, name_map)
        if columns != columns_previous:
            raise ValueError('files have different columns')



"""Analysis for accretion discs."""

import numpy as np
from numpy import ndarray

from ..snap.snap import Snap
from .total import angular_momentum, center_of_mass


def normal(snap: Snap, ignore_accreted: bool = True) -> ndarray:
    """Calculate unit normal to disc.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    ndarray
        A unit normal to the plane of the disc.
    """
    origin = center_of_mass(snap=snap, ignore_accreted=ignore_accreted)
    L = angular_momentum(
        snap=snap, origin=origin, ignore_accreted=ignore_accreted
    ).magnitude
    return L / np.linalg.norm(L)


def rotate_face_on(snap: Snap, ignore_accreted: bool = True) -> Snap:
    """Rotate disc to face-on.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    Snap
        The rotated Snap.
    """
    x, y, z = normal(snap=snap, ignore_accreted=ignore_accreted)
    axis = (-x, y, 0)
    angle = np.arctan(np.sqrt(x ** 2 + y ** 2) / z)
    return snap.rotate(axis=axis, angle=angle)


def rotate_edge_on(snap: Snap, ignore_accreted: bool = True) -> Snap:
    """Rotate disc to edge-on.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    Snap
        The rotated Snap.
    """
    snap = rotate_face_on(snap=snap, ignore_accreted=ignore_accreted)
    return snap.rotate(axis=(1, 0, 0), angle=np.pi / 2)



"""Test Profile."""

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import plonk

TEST_FILE = Path(__file__).parent / 'stubdata/phantom_00000.h5'
CSV_FILE = Path(__file__).parent / 'stubdata/phantom_00000_profile.csv'


def test_load_profile():
    """Test loading Profile."""
    snap = plonk.load_snap(TEST_FILE)

    prof = plonk.load_profile(snap=snap)
    for p in ['aspect_ratio', 'angular_momentum_phi', 'angular_momentum_theta']:
        prof[p]
    with pytest.raises(ValueError):
        prof['does_not_exist']
    plonk.load_profile(snap=snap, ndim=3, cmin='10 au', cmax='100 au', n_bins=30)
    plonk.load_profile(snap=snap, spacing='log', ignore_accreted=False)
    plonk.load_profile(snap=snap, ndim=1, coordinate='x')
    plonk.load_profile(snap=snap, ndim=1, coordinate='y')
    plonk.load_profile(snap=snap, ndim=1, coordinate='z')
    with pytest.raises(ValueError):
        plonk.load_profile(snap=snap, ndim=1, coordinate='does_not_exist')
    with pytest.raises(ValueError):
        plonk.load_profile(snap=snap, cmin=10, cmax=100)

    snap.close_file()


def test_check_data():
    """Test Profile data accuracy."""
    snap = plonk.load_snap(TEST_FILE)
    prof = plonk.load_profile(snap)

    columns = [
        'density',
        'mass',
        'number',
        'radius',
        'scale_height',
        'size',
        'smoothing_length',
        'sound_speed',
    ]

    df = prof.to_dataframe(columns=columns)

    units = ['g/cm^3', 'g', '', 'au', 'au', 'au^2', 'au', 'km/s']
    df = prof.to_dataframe(columns=columns, units=units)
    pd.testing.assert_frame_equal(df, pd.read_csv(CSV_FILE, index_col=0))

    snap.close_file()


def test_set_data():
    """Test setting array on Profile."""
    snap = plonk.load_snap(TEST_FILE)

    prof = plonk.load_profile(snap=snap)
    prof['array'] = np.arange(len(prof)) * plonk.units.au
    with pytest.raises(ValueError):
        prof['array'] = np.arange(len(prof) - 1)
    with pytest.raises(ValueError):
        prof['array'] = 1.0

    snap.close_file()


def test_profile_plot():
    """Test loading Profile."""
    snap = plonk.load_snap(TEST_FILE)

    prof = plonk.load_profile(snap=snap)
    prof.plot(x='radius', y='surface_density')
    prof.plot(
        x='radius',
        y='density',
        units={'position': 'au', 'density': 'g/cm^3'},
        std='shading',
    )

    snap.close_file()


def test_to_function():
    """Test to_function."""
    snap = plonk.load_snap(TEST_FILE)
    prof = plonk.load_profile(snap=snap)

    fn = prof.to_function(profile='scale_height')
    assert np.allclose(fn(prof['radius']), prof['scale_height'])
    snap.close_file()



"""Analysis of SPH data.

The analysis sub-package contains Plonk implementations of typical
smoothed particle hydrodynamics post-simulation analysis tasks.

Examples
--------
Create a radial profile in the xy-plane.

>>> p = Profile(snap, cmin=10, cmax=200, n_bins=100)
>>> p.plot('radius', 'density')

Calculate the angular momentum on the particles.

>>> angmom = particles.angular_momentum(snap)

Calculate the total angular momentum over all particles.

>>> angmom_tot = total.angular_momentum(snap)

Calculate the Roche sphere radius given two sink particles.

>>> s1 = snap.sinks[0]
>>> s2 = snap.sinks[1]
>>> separation = plonk.utils.math.norm(s1['position'] - s2['position'])
>>> Roche = sinks.Roche_sphere(s1['mass'], s2['mass'], separation)
"""

from . import discs, filters, particles, sinks, sph, total
from .profile import Profile, load_profile

__all__ = [
    'Profile',
    'discs',
    'filters',
    'load_profile',
    'particles',
    'sinks',
    'sph',
    'total',
]



"""Utility functions."""


def is_documented_by(original):
    """Wrap function to add docstring."""

    def wrapper(target):
        target.__doc__ = original.__doc__
        return target

    return wrapper


def time_string(
    snap, unit: str, unit_str: str = None, float_format: str = '.0f',
) -> str:
    """Generate time stamp string.

    Parameters
    ----------
    snap
        The Snap object.
    unit
        The time unit as a string to pass to Pint. E.g. 'year'.
    unit_str
        The unit string to print. If None, use the value from 'unit'.
        Default is None.
    float_format
        The format for the time float value. Default is '.0f'.

    Examples
    --------
    Generate a list of strings of snapshot time, like
    ['0 yr', '10 yr', ...].

    >>> text = [time_string(snap, 'year', 'yr') for snap in snaps]

    Or, in terms of an orbital time, like '10 orbits'.

    >>> plonk.units.define('binary_orbit = 100 years')
    >>> time_string(snap, 'binary_orbit', 'orbits')
    """
    time = snap.properties['time'].to(unit).magnitude
    if unit_str is None:
        unit_str = unit
    return f't = {time:{float_format}} {unit_str}'


def pretty_array_name(s: str) -> str:
    """Prettify an array name string.

    Parameters
    ----------
    s
        The string.

    Returns
    -------
    str
    """
    NUM_DUST_MAX = 100

    for v in ['x', 'y', 'z', 'R', 'r', 'phi', 'theta']:
        if s.endswith(f'_{v}'):
            s = _rreplace(s, f'_{v}', f' ({v})', 1)
    for v in [f'{n:03}' for n in range(NUM_DUST_MAX)]:
        if s.endswith(f'_{v}'):
            s = _rreplace(s, f'_{v}', f' ({v})', 1)
    if s.endswith('_tot'):
        s = _rreplace(s, '_tot', ' (total)', 1)
    if s.endswith('_mag'):
        s = _rreplace(s, '_mag', ' (magnitude)', 1)
    if len(s) > 1:
        s = s.capitalize()
    return s.replace('_', ' ')


def _rreplace(s, old, new, occurrence):
    # From https://stackoverflow.com/a/2556252/12951362
    li = s.rsplit(old, occurrence)
    return new.join(li)



"""Calculate global (total) quantities on the particles."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from .._units import Quantity
from ..utils.math import norm
from . import particles

if TYPE_CHECKING:
    from ..snap.snap import SnapLike


def accreted_mass(snap: SnapLike) -> float:
    """Calculate the accreted mass.

    Parameters
    ----------
    snap
        The Snap object.

    Returns
    -------
    float
        The accreted mass.
    """
    h: Quantity = snap['smoothing_length']
    _mass: Quantity = snap['mass'][~(h > 0)]

    return _mass.sum()


def angular_momentum(
    snap: SnapLike, origin: Quantity = None, ignore_accreted: bool = True,
) -> Quantity:
    """Calculate the total angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular momentum as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    Quantity
        The total angular momentum like (lx, ly, lz).
    """
    return particles.angular_momentum(
        snap=snap, origin=origin, ignore_accreted=ignore_accreted
    ).sum(axis=0)


def center_of_mass(snap: SnapLike, ignore_accreted: bool = True) -> Quantity:
    """Calculate the center of mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    Quantity
        The center of mass as a vector (cx, cy, cz).
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        _mass: Quantity = snap['mass'][h > 0]
        pos: Quantity = snap['position'][h > 0]
    else:
        _mass = snap['mass']
        pos = snap['position']

    return (_mass[:, np.newaxis] * pos).sum(axis=0) / _mass.sum()


def dust_mass(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total dust mass per species.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total dust mass per species.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        _mass: Quantity = snap['mass'][h > 0]
        dustfrac: Quantity = snap['dustfrac'][h > 0]
    else:
        _mass = snap['mass']
        dustfrac = snap['dustfrac']

    return (_mass[:, np.newaxis] * dustfrac).sum(axis=0)


def gas_mass(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total gas mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total gas mass.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        _mass: Quantity = snap['mass'][h > 0]
        dustfrac: Quantity = snap['dustfrac'][h > 0]
    else:
        _mass = snap['mass']
        dustfrac = snap['dustfrac']

    gas_frac = 1 - dustfrac.sum(axis=1)
    return (_mass * gas_frac).sum()


def inclination(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the inclination with respect to the xy-plane.

    The inclination is calculated by taking the angle between the
    angular momentum vector and the z-axis, with the angular momentum
    calculated with respect to the center of mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The mean inclination.
    """
    angmom = angular_momentum(snap=snap, ignore_accreted=ignore_accreted)
    return np.arccos(angmom[2] / norm(angmom))


def kinetic_energy(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total kinetic energy.
    """
    return particles.kinetic_energy(snap=snap, ignore_accreted=ignore_accreted).sum(
        axis=0
    )


def mass(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total mass.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        _mass: Quantity = snap['mass'][h > 0]
    else:
        _mass = snap['mass']

    return _mass.sum()


def momentum(snap: SnapLike, ignore_accreted: bool = True) -> Quantity:
    """Calculate the total momentum.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    Quantity
        The total linear momentum like (px, py, pz).
    """
    return particles.momentum(snap=snap, ignore_accreted=ignore_accreted).sum(axis=0)


def position_angle(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the position angle of inclination.

    The position angle is taken from the x-axis in the xy-plane. It
    defines a unit vector around which the snap is inclined.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The mean inclination.
    """
    angmom = angular_momentum(snap=snap, ignore_accreted=ignore_accreted)
    if isinstance(angmom, Quantity):
        pi_2 = np.pi / 2 * Quantity('radian')
    else:
        pi_2 = np.pi / 2
    return np.arctan2(angmom[1], angmom[0]) + pi_2


def specific_angular_momentum(
    snap: SnapLike, origin: Quantity = None, ignore_accreted: bool = True,
) -> Quantity:
    """Calculate the total specific angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular momentum as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    Quantity
        The total specific angular momentum on the particles like
        (hx, hy, hz).
    """
    return particles.specific_angular_momentum(
        snap=snap, origin=origin, ignore_accreted=ignore_accreted
    ).sum(axis=0)


def specific_kinetic_energy(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total specific kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total specific kinetic energy.
    """
    return particles.specific_kinetic_energy(
        snap=snap, ignore_accreted=ignore_accreted
    ).sum(axis=0)



"""Physical units on snapshots."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Union

from .._config import load_config
from .._units import units as plonk_units

if TYPE_CHECKING:
    from .snap import SnapLike


def gravitational_constant_in_code_units(snap: SnapLike) -> float:
    """Gravitational constant in code units.

    Parameters
    ----------
    snap
        The Snap object.

    Returns
    -------
    float
        The gravitational constant in code units.
    """
    G = plonk_units.newtonian_constant_of_gravitation
    G_units = (
        snap.code_units['length'] ** 3
        / snap.code_units['mass']
        / snap.code_units['time'] ** 2
    )
    G = (G / G_units).to_base_units().magnitude
    return G


def add_aliases(snap: SnapLike, filename: Union[str, Path] = None):
    """Add array aliases to a Snap.

    Parameters
    ----------
    snap
        The Snap object.
    config : optional
        The path to a Plonk config.toml file. If None, use the default
        file.
    """
    conf = load_config(filename=filename)
    for key, val in conf['arrays']['aliases'].items():
        snap.add_alias(key, val)



"""Splash interpolation functions.

Derived from Splash: https://github.com/danieljprice/splash.
"""

import numba
import numpy as np
from numpy import ndarray

RADKERNEL = 2.0
RADKERNEL2 = 4.0
CNORMK3D = 1.0 / np.pi

NPTS = 100
MAXCOLTABLE = 1000
DQ2TABLE = RADKERNEL2 / MAXCOLTABLE
DDQ2TABLE = 1.0 / DQ2TABLE

IVERBOSE = -1


@numba.njit
def w_cubic(q2: float):
    """Cubic spline kernel.

    Parameters
    ----------
    q2

    Returns
    -------
    w
    """
    w = 0.0
    if q2 < 1.0:
        q = np.sqrt(q2)
        w = 1.0 - 1.5 * q2 + 0.75 * q2 * q
    elif q2 < 4:
        q = np.sqrt(q2)
        w = 0.25 * (2.0 - q) ** 3
    return w


@numba.njit
def setup_integratedkernel():
    """Set up integrated kernel.

    Tabulates the integral through the cubic spline kernel tabulated in
    (r/h)**2 so that sqrt is not necessary.

    Returns
    -------
    coltable
    """
    coltable = np.zeros(MAXCOLTABLE + 1)

    for idx in range(MAXCOLTABLE):
        # Tabulate for (cylindrical) r**2 between 0 and RADKERNEL**2
        rxy2 = idx * DQ2TABLE

        # Integrate z between 0 and sqrt(RADKERNEL^2 - rxy^2)
        deltaz = np.sqrt(RADKERNEL2 - rxy2)
        dz = deltaz / (NPTS - 1)
        coldens = 0
        for j in range(NPTS):
            z = j * dz
            q2 = rxy2 + z * z
            wkern = w_cubic(q2)
            if j in (0, NPTS - 1):
                coldens = coldens + 0.5 * wkern * dz
            else:
                coldens = coldens + wkern * dz
        coltable[idx] = 2.0 * coldens * CNORMK3D
    coltable[MAXCOLTABLE] = 0.0

    return coltable


@numba.njit
def wfromtable(q2, coltable):
    """Interpolate from integrated kernel table values to give w(q).

    Parameters
    ----------
    q2
    coltable

    Returns
    -------
    w
    """
    # Find nearest index in table
    index = int(q2 * DDQ2TABLE)
    index1 = min(index, MAXCOLTABLE)

    # Find increment along from this index
    dxx = q2 - index * DQ2TABLE

    # Find gradient
    dwdx = (coltable[index1] - coltable[index]) * DDQ2TABLE

    # Compute value of integrated kernel
    return coltable[index] + dwdx * dxx


@numba.njit
def interpolate_projection(
    x: ndarray,
    y: ndarray,
    hh: ndarray,
    weight: ndarray,
    dat: ndarray,
    itype: ndarray,
    npart: int,
    xmin: float,
    ymin: float,
    npixx: int,
    npixy: int,
    pixwidthx: float,
    pixwidthy: float,
    normalise: bool,
):
    """Interpolate particles to grid via projection.

    Parameters
    ----------
    x
        The particle x positions.
    y
        The particle y positions.
    hh
        The particle smoothing length.
    weight
        The particle weight.
    dat
        The scalar data to interpolate.
    itype
        The particle type.
    npart
        The number of particles.
    xmin
        The minimum x position.
    ymin
        The minimum y position.
    npixx
        The number of pixels in the x direction.
    npixy
        The number of pixels in the y direction.
    normalise
        Whether to normalize.

    Return
    ------
    datsmooth
        The data smoothed to a pixel grid.
    """
    coltable = setup_integratedkernel()

    datsmooth = np.zeros((npixx, npixy))
    datnorm = np.zeros((npixx, npixy))
    dx2i = np.zeros(npixx)
    xpix = np.zeros(npixx)
    term = 0.0

    xminpix = xmin - 0.5 * pixwidthx
    yminpix = ymin - 0.5 * pixwidthy
    xmax = xmin + npixx * pixwidthx
    ymax = ymin + npixy * pixwidthy

    # Use a minimum smoothing length on the grid to make sure that particles
    # contribute to at least one pixel
    hmin = 0.5 * max(pixwidthx, pixwidthy)

    xpix = xminpix + np.arange(1, npixx + 1) * pixwidthx
    nsubgrid = 0
    nok = 0
    hminall = 1e10

    # Loop over particles
    for idx in range(npart):

        # Skip particles with itype < 0
        if itype[idx] < 0:
            continue

        # Set h related quantities
        hi = hh[idx]
        horigi = hi
        if not hi > 0.0:
            continue

        # Radius of the smoothing kernel
        radkern = RADKERNEL * hi

        # Cycle as soon as we know the particle does not contribute
        xi = x[idx]
        xpixmin = xi - radkern
        if xpixmin > xmax:
            continue
        xpixmax = xi + radkern
        if xpixmax < xmin:
            continue

        yi = y[idx]
        ypixmin = yi - radkern
        if ypixmin > ymax:
            continue
        ypixmax = yi + radkern
        if ypixmax < ymin:
            continue

        # Take resolution length as max of h and 1/2 pixel width
        if hi < hmin:
            hminall = min(hi, hminall)
            nsubgrid = nsubgrid + 1
            hsmooth = hmin
        else:
            hsmooth = hi
            nok = nok + 1
        radkern = RADKERNEL * hsmooth

        # Set kernel related quantities
        hi1 = 1.0 / hsmooth
        hi21 = hi1 * hi1
        termnorm = weight[idx] * horigi
        term = termnorm * dat[idx]

        # Loop over pixels, adding the contribution from this particle
        # copy by quarters if all pixels within domain
        ipixmin = int((xi - radkern - xmin) / pixwidthx)
        ipixmax = int((xi + radkern - xmin) / pixwidthx) + 1
        jpixmin = int((yi - radkern - ymin) / pixwidthy)
        jpixmax = int((yi + radkern - ymin) / pixwidthy) + 1

        # Make sure they only contribute to pixels in the image
        # (note that this optimises much better than using min/max)
        if ipixmin < 0:
            ipixmin = 0
        if jpixmin < 0:
            jpixmin = 0
        if ipixmax > npixx:
            ipixmax = npixx
        if jpixmax > npixy:
            jpixmax = npixy

        # Precalculate an array of dx2 for this particle (optimisation)
        for ipix in range(ipixmin, ipixmax):
            dx2i[ipix] = ((xpix[ipix] - xi) ** 2) * hi21

        for jpix in range(jpixmin, jpixmax):
            ypix = yminpix + jpix * pixwidthy
            dy = ypix - yi
            dy2 = dy * dy * hi21
            for ipix in range(ipixmin, ipixmax):
                # dx2 pre-calculated; dy2 pre-multiplied by hi21
                q2 = dx2i[ipix] + dy2
                # SPH kernel - integral through cubic spline
                # interpolate from a pre-calculated table
                if q2 < RADKERNEL2:
                    wab = wfromtable(q2, coltable)
                    # Calculate data value at this pixel using the summation interpolant
                    datsmooth[ipix, jpix] = datsmooth[ipix, jpix] + term * wab
                    if normalise:
                        datnorm[ipix, jpix] = datnorm[ipix, jpix] + termnorm * wab

    # Normalise dat array
    if normalise:
        # Normalise everywhere (required if not using SPH weighting)
        for idxi in range(npixx):
            for idxj in range(npixy):
                if datnorm[idxi, idxj] > 0.0:
                    datsmooth[idxi, idxj] /= datnorm[idxi, idxj]

    # Warn about subgrid interpolation
    if nsubgrid > 1:
        nfull = int((xmax - xmin) / (hminall)) + 1
        if nsubgrid > 0.1 * nok and IVERBOSE > -1:
            print('Warning: pixel size > 2h for ', nsubgrid, ' particles')
            print('need ', nfull, ' pixels for full resolution')

    # Return datsmooth
    return datsmooth.T


@numba.njit
def interpolate_slice(
    x: ndarray,
    y: ndarray,
    dslice: ndarray,
    hh: ndarray,
    weight: ndarray,
    dat: ndarray,
    itype: ndarray,
    npart: int,
    xmin: float,
    ymin: float,
    npixx: int,
    npixy: int,
    pixwidthx: float,
    pixwidthy: float,
    normalise: bool,
):
    """Interpolate particles to grid via cross section.

    Parameters
    ----------
    x
        The particle x positions.
    y
        The particle y positions.
    dslice
        The distance from the cross section slice.
    hh
        The particle smoothing length.
    weight
        The particle weight.
    dat
        The scalar data to interpolate.
    itype
        The particle type.
    npart
        The number of particles.
    xmin
        The minimum x position.
    ymin
        The minimum y position.
    zslice
        Cross section location.
    npixx
        The number of pixels in the x direction.
    npixy
        The number of pixels in the y direction.
    normalise
        Whether to normalize.

    Return
    ------
    datsmooth
        The data smoothed to a pixel grid.
    """
    datsmooth = np.zeros((npixx, npixy))
    datnorm = np.zeros((npixx, npixy))
    dx2i = np.zeros(npixx)
    const = CNORMK3D

    # Loop over particles
    for idx in range(npart):

        # Skip particles with itype < 0
        if itype[idx] < 0:
            continue

        # Set h related quantities
        hi = hh[idx]
        if not hi > 0.0:
            continue
        hi1 = 1.0 / hi
        hi21 = hi1 * hi1
        radkern = RADKERNEL * hi

        # For each particle, work out distance from the cross section slice
        dz2 = dslice[idx] ** 2 * hi21

        # If this is < 2h then add the particle's contribution to the pixels
        # otherwise skip all this and start on the next particle
        if dz2 < RADKERNEL2:

            xi = x[idx]
            yi = y[idx]
            termnorm = const * weight[idx]
            term = termnorm * dat[idx]

            # Loop over pixels, adding the contribution from this particle
            # copy by quarters if all pixels within domain
            ipixmin = int((xi - radkern - xmin) / pixwidthx)
            ipixmax = int((xi + radkern - xmin) / pixwidthx) + 1
            jpixmin = int((yi - radkern - ymin) / pixwidthy)
            jpixmax = int((yi + radkern - ymin) / pixwidthy) + 1

            # Make sure they only contribute to pixels in the image
            # (note that this optimises much better than using min/max)
            if ipixmin < 0:
                ipixmin = 0
            if jpixmin < 0:
                jpixmin = 0
            if ipixmax > npixx:
                ipixmax = npixx
            if jpixmax > npixy:
                jpixmax = npixy

            # Precalculate an array of dx2 for this particle (optimisation)
            for ipix in range(ipixmin, ipixmax):
                dx2i[ipix] = ((xmin + (ipix - 0.5) * pixwidthx - xi) ** 2) * hi21 + dz2

            # Loop over pixels, adding the contribution from this particle
            for jpix in range(jpixmin, jpixmax):
                ypix = ymin + (jpix - 0.5) * pixwidthy
                dy = ypix - yi
                dy2 = dy * dy * hi21
                for ipix in range(ipixmin, ipixmax):
                    q2 = dx2i[ipix] + dy2
                    # SPH kernel - cubic spline
                    if q2 < RADKERNEL2:
                        wab = w_cubic(q2)
                        # Calculate data value at this pixel using the summation
                        # interpolant
                        datsmooth[ipix, jpix] = datsmooth[ipix, jpix] + term * wab
                        if normalise:
                            datnorm[ipix, jpix] = datnorm[ipix, jpix] + termnorm * wab

    # Normalise dat array
    if normalise:
        # Normalise everywhere (required if not using SPH weighting)
        for idxi in range(npixx):
            for idxj in range(npixy):
                if datnorm[idxi, idxj] > 0.0:
                    datsmooth[idxi, idxj] /= datnorm[idxi, idxj]

    # Return datsmooth
    return datsmooth.T



"""Functions for visualization."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection
from matplotlib.patches import Circle

from .._units import Quantity

if TYPE_CHECKING:
    from ..snap.snap import SnapLike


def plot_smoothing_length(
    snap: SnapLike,
    indices: List[int],
    fac: float = 1.0,
    units: Union[str, Quantity] = None,
    x: str = 'x',
    y: str = 'y',
    ax: Any = None,
    **kwargs
) -> Any:
    """Plot smoothing length around particle.

    Also works for plotting the accretion radius on a sink particle.

    Parameters
    ----------
    snap
        The Snap object.
    indices
        The particle indices.
    fac
        Set the circle to be a multiple "fac" of the smoothing length.
    units
        The distance units.
    ax
        The matplotlib Axes to plot on.
    x
        The "x" coordinate.
    y
        The "y" coordinate.
    **kwargs
        Keyword arguments to pass to matplotlib PatchCollection.

    Returns
    -------
    circles
        A list of matplotlib circles.
    """
    px: Quantity = snap[x][indices]
    py: Quantity = snap[y][indices]
    try:
        h: Quantity = snap['smoothing_length'][indices]
    except (ValueError, KeyError):
        # Sinks accretion radius is equivalent to smoothing length
        h = snap['accretion_radius'][indices]
    if units is None:
        arr: Quantity = snap[x]
        units = arr.units
    px = px.to(units).magnitude
    py = py.to(units).magnitude
    h = h.to(units).magnitude

    circles = [Circle((_px, _py), fac * _h) for _px, _py, _h in zip(px, py, h)]
    collection = PatchCollection(circles, **kwargs)
    if ax is None:
        fig, ax = plt.subplots()
        ax.set_aspect('equal')
    ax.add_collection(collection)
    _set_new_lim(ax, px, py, h, fac)

    return collection


def get_extent_from_percentile(
    snap: SnapLike,
    x: str,
    y: str,
    percentile: float = 99,
    x_center_on: float = None,
    y_center_on: float = None,
    edge_factor: float = None,
):
    """Get extent from percentile.

    Parameters
    ----------
    snap
        The Snap object.
    x
        The "x" coordinate.
    y
        The "y" coordinate.
    percentile : optional
        The percentile used in the calculation. Default is 99.
    x_center_on : optional
        Center on some x-value. Default is None.
    y_center_on : optional
        Center on some y-value. Default is None.
    edge_factor : optional
        Add extra spacing to extent. E.g. to add extra 5%, set this
        value to 0.05. Default is None.

    Returns
    -------
    tuple
        The extent of the box as (xmin, xmax, ymin, ymax).
    """
    pl, pr = (100 - percentile) / 2, percentile + (100 - percentile) / 2
    xlim = np.percentile(snap[x], [pl, pr])
    ylim = np.percentile(snap[y], [pl, pr])

    if x_center_on is not None:
        xlim += x_center_on - xlim.mean()
    if y_center_on is not None:
        ylim += y_center_on - ylim.mean()

    if edge_factor is not None:
        dx = xlim[1] - xlim[0]
        dy = ylim[1] - ylim[0]
        xlim += (-dx * edge_factor, dx * edge_factor)
        ylim += (-dy * edge_factor, dy * edge_factor)
        return (xlim[0], xlim[1], ylim[0], ylim[1])

    return (xlim[0], xlim[1], ylim[0], ylim[1])


def _set_new_lim(ax, x, y, h, fac):
    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    x_min = x.min() - fac * h.max()
    x_max = x.max() + fac * h.max()
    y_min = y.min() - fac * h.max()
    y_max = y.max() + fac * h.max()
    _xlim = (min(x_min, xlim[0]), max(x_max, xlim[1]))
    _ylim = (min(y_min, ylim[0]), max(y_max, ylim[1]))
    ax.set_xlim(_xlim)
    ax.set_ylim(_ylim)



"""Calculate extra quantities on the particles."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from .._logging import logger
from .._units import Quantity
from .._units import units as plonk_units
from ..utils.math import cross, norm

if TYPE_CHECKING:
    from ..snap.snap import SnapLike

ORIGIN = (0, 0, 0) * plonk_units.au
MOLECULAR_HYDROGEN_WEIGHT = 2.381

# Derived quantities require some arrays already present
array_requires = {
    'angular_momentum': ('mass', 'position', 'velocity'),
    'angular_velocity': ('position', 'velocity'),
    'azimuthal_angle': ('position',),
    'dust_density': ('dust_fraction',),
    'dust_fraction': ('dust_fraction',),
    'dust_mass': ('dust_fraction',),
    'eccentricity': ('position', 'velocity'),
    'gas_density': ('dust_fraction',),
    'gas_fraction': ('dust_fraction',),
    'gas_mass': ('dust_fraction',),
    'inclination': ('mass', 'position', 'velocity'),
    'keplerian_frequency': ('position',),
    'kinetic_energy': ('mass', 'velocity'),
    'momentum': ('mass', 'velocity'),
    'polar_angle': ('position',),
    'radius_cylindrical': ('position',),
    'radius_spherical': ('position',),
    'semi_major_axis': ('position', 'velocity'),
    'specific_angular_momentum': ('position', 'velocity'),
    'specific_kinetic_energy': 'velocity',
    'stokes_number': ('position', 'stopping_time'),
    'temperature': ('sound_speed',),
    'velocity_radial_cylindrical': ('position', 'velocity'),
    'velocity_radial_spherical': ('position', 'velocity'),
}

# Arrays which represent quantities with x-, y-, z-components in space
vector_arrays = ['angular_momentum', 'momentum', 'specific_angular_momentum']

# Arrays which represent dust quantities with columns per dust species
dust_arrays = ['dust_density', 'dust_fraction', 'dust_mass', 'stokes_number']


def angular_momentum(
    snap: SnapLike, origin: Quantity = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular momentum as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The angular momentum on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        mass: Quantity = snap['mass'][h > 0]
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        pos = snap['position']
        vel = snap['velocity']

    origin = snap.translation if snap.translation is not None else ORIGIN
    pos = pos - origin

    return mass[:, np.newaxis] * cross(pos, vel)


def angular_velocity(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the angular velocity.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The angular velocity on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    x, y = pos[:, 0], pos[:, 1]
    vx, vy = vel[:, 0], vel[:, 1]

    vphi = (x * vy - y * vx) / (x ** 2 + y ** 2)

    return vphi


def azimuthal_angle(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the azimuthal angle.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The azimuthal angle on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    x, y = pos[:, 0], pos[:, 1]
    return np.arctan2(y, x)


def dust_density(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the dust density per species.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The dust density per species on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        n_dust = len(snap.properties.get('grain_size', []))
        sub_type = snap['sub_type']
        type_mask = snap['type'] == snap.particle_type['dust']
        _dust_fraction = np.zeros((len(snap), n_dust))
        for idx in range(n_dust):
            _dust_fraction[type_mask & (sub_type == idx + 1), idx] = 1

    else:
        raise ValueError('No dust available')

    density: Quantity = snap['density']
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        _dust_mass = density[:, np.newaxis] * _dust_fraction
        return _dust_mass[h > 0]
    return density[:, np.newaxis] * _dust_fraction


def dust_fraction(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the dust fraction.

    For dust/gas mixtures this exists already.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The gas fraction on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        n_dust = len(snap.properties.get('grain_size', []))
        sub_type = snap['sub_type']
        type_mask = snap['type'] == snap.particle_type['dust']
        _dust_fraction = np.zeros((len(snap), n_dust))
        for idx in range(n_dust):
            _dust_fraction[type_mask & (sub_type == idx + 1), idx] = 1

    else:
        raise ValueError('No dust available')

    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        return _dust_fraction[h > 0]
    return _dust_fraction


def dust_mass(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the dust mass per species.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The dust mass per species on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        n_dust = len(snap.properties.get('grain_size', []))
        sub_type = snap['sub_type']
        type_mask = snap['type'] == snap.particle_type['dust']
        _dust_fraction = np.zeros((len(snap), n_dust))
        for idx in range(n_dust):
            _dust_fraction[type_mask & (sub_type == idx + 1), idx] = 1

    else:
        raise ValueError('No dust available')

    mass: Quantity = snap['mass']
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        _dust_mass = mass[:, np.newaxis] * _dust_fraction
        return _dust_mass[h > 0]
    return mass[:, np.newaxis] * _dust_fraction


def eccentricity(
    snap: SnapLike,
    gravitational_parameter: Quantity = None,
    origin: Quantity = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the eccentricity.

    The eccentricity of particles around a mass specified by
    gravitational parameter with an optional to specify the position of
    the mass.

    Parameters
    ----------
    snap
        The Snap object.
    gravitational_parameter
        The gravitational parameter (mu = G M) as a Pint quantity.
    origin : optional
        The origin around which to compute the eccentricity as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The eccentricity on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    origin = snap.translation if snap.translation is not None else ORIGIN
    pos = pos - origin

    mu = snap.properties.get('gravitational_parameter')
    if mu is None:
        raise ValueError(
            'must pass in gravitational_parameter or '
            'set on Snap with set_gravitational_parameter'
        )

    radius = norm(pos, axis=1)

    _specific_angular_momentum = cross(pos, vel)
    specific_angular_momentum_magnitude = norm(_specific_angular_momentum, axis=1)

    _specific_kinetic_energy = 1 / 2 * norm(vel, axis=1) ** 2
    specific_potential_energy = -mu / radius
    specific_energy = _specific_kinetic_energy + specific_potential_energy

    term = specific_energy * (specific_angular_momentum_magnitude / mu) ** 2
    return np.sqrt(1 + 2 * term)


def gas_density(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the gas density.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The gas density on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']
        _gas_fraction = 1 - _dust_fraction.sum(axis=1)

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        particle_type = snap['type']
        _gas_fraction = np.zeros(len(snap))
        _gas_fraction[particle_type == snap.particle_type['gas']] = 1

    density: Quantity = snap['density']
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        return (density * _gas_fraction)[h > 0]
    return density * _gas_fraction


def gas_fraction(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the gas fraction.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The gas fraction on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']
        _gas_fraction = 1 - _dust_fraction.sum(axis=1)

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        particle_type = snap['type']
        _gas_fraction = np.ones(len(snap))
        _gas_fraction[particle_type != snap.particle_type['gas']] = 0

    else:
        _gas_fraction = np.ones(len(snap))

    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        return _gas_fraction[h > 0]
    return _gas_fraction


def gas_mass(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the gas mass.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The gas mass on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']
        _gas_fraction = 1 - _dust_fraction.sum(axis=1)

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        particle_type = snap['type']
        _gas_fraction = np.zeros(len(snap))
        _gas_fraction[particle_type == snap.particle_type['gas']] = 1

    mass: Quantity = snap['mass']
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        return (mass * _gas_fraction)[h > 0]
    return mass * _gas_fraction


def inclination(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the inclination.

    The inclination is calculated by taking the angle between the
    angular momentum vector and the z-axis, with the angular momentum
    calculated with respect to the center of mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The inclination on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        mass: Quantity = snap['mass'][h > 0]
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        pos = snap['position']
        vel = snap['velocity']

    origin = (mass[:, np.newaxis] * pos).sum(axis=0) / mass.sum()
    pos = pos - origin

    _specific_angular_momentum = cross(pos, vel)

    return np.arccos(
        _specific_angular_momentum[:, 2] / norm(_specific_angular_momentum, axis=1)
    )


def keplerian_frequency(
    snap: SnapLike,
    gravitational_parameter: Quantity = None,
    origin: Quantity = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the Keplerian orbital frequency.

    The Keplerian orbital frequency of particles around a mass specified
    by gravitational parameter with an optional to specify the position
    of the mass.

    Parameters
    ----------
    snap
        The Snap object.
    gravitational_parameter
        The gravitational parameter (mu = G M) as a Pint quantity.
    origin : optional
        The origin around which to compute the Keplerian frequency as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The Keplerian frequency on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    origin = snap.translation if snap.translation is not None else ORIGIN
    pos = pos - origin

    mu = snap.properties.get('gravitational_parameter')
    if mu is None:
        raise ValueError(
            'must pass in gravitational_parameter or '
            'set on Snap with set_gravitational_parameter'
        )

    radius = norm(pos, axis=1)
    return np.sqrt(mu / radius ** 3)


def kinetic_energy(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The kinetic energy on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        mass: Quantity = snap['mass'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        vel = snap['velocity']

    return 1 / 2 * mass * norm(vel, axis=1) ** 2


def momentum(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the momentum.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The linear momentum on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        mass: Quantity = snap['mass'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        vel = snap['velocity']

    return mass[:, np.newaxis] * vel


def polar_angle(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the polar angle.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The azimuthal angle on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]
    r = np.sqrt(x ** 2 + y ** 2 + z ** 2)
    return np.arccos(z / r)


def radius_cylindrical(
    snap: SnapLike, origin: Quantity = None, ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the cylindrical radial distance.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The radial distance on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    origin = snap.translation if snap.translation is not None else ORIGIN
    pos = pos - origin
    x, y = pos[:, 0], pos[:, 1]

    return np.sqrt(x ** 2 + y ** 2)


def radius_spherical(
    snap: SnapLike, origin: Quantity = None, ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the spherical radial distance.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The radial distance on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    origin = snap.translation if snap.translation is not None else ORIGIN
    pos = pos - origin
    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]

    return np.sqrt(x ** 2 + y ** 2 + z ** 2)


def semi_major_axis(
    snap: SnapLike,
    gravitational_parameter: Quantity = None,
    origin: Quantity = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the semi-major axis.

    The semi-major axis of particles around a mass specified by
    gravitational parameter with an optional to specify the position of
    the mass.

    Parameters
    ----------
    snap
        The Snap object.
    gravitational_parameter
        The gravitational parameter (mu = G M) as a Pint quantity.
    origin : optional
        The origin around which to compute the semi-major axis as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The semi-major axis on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    origin = snap.translation if snap.translation is not None else ORIGIN
    pos = pos - origin

    mu = snap.properties.get('gravitational_parameter')
    if mu is None:
        raise ValueError(
            'must pass in gravitational_parameter or '
            'set on Snap with set_gravitational_parameter'
        )

    radius = norm(pos, axis=1)

    _specific_angular_momentum = cross(pos, vel)
    specific_angular_momentum_magnitude = norm(_specific_angular_momentum, axis=1)

    _specific_kinetic_energy = 1 / 2 * norm(vel, axis=1) ** 2
    specific_potential_energy = -mu / radius
    specific_energy = _specific_kinetic_energy + specific_potential_energy

    term = specific_energy * (specific_angular_momentum_magnitude / mu) ** 2

    _eccentricity = np.sqrt(1 + 2 * term)

    return specific_angular_momentum_magnitude ** 2 / (mu * (1 - _eccentricity ** 2))


def specific_angular_momentum(
    snap: SnapLike, origin: Quantity = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the specific angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular momentum as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The specific angular momentum on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    origin = snap.translation if snap.translation is not None else ORIGIN
    pos = pos - origin

    return cross(pos, vel)


def specific_kinetic_energy(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the specific kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The specific kinetic energy on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        vel: Quantity = snap['velocity'][h > 0]
    else:
        vel = snap['velocity']

    return 1 / 2 * norm(vel, axis=1) ** 2


def stokes_number(
    snap: SnapLike,
    gravitational_parameter: Quantity = None,
    origin: Quantity = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the Stokes number.

    Parameters
    ----------
    snap
        The Snap object.
    gravitational_parameter
        The gravitational parameter (mu = G M) as a Pint quantity.
    origin : optional
        The origin around which to compute the Stokes number as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The Stokes number on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        t_s: Quantity = snap['stopping_time'][h > 0]
    else:
        pos = snap['position']
        t_s = snap['stopping_time']

    origin = snap.translation if snap.translation is not None else ORIGIN
    pos = pos - origin

    mu = snap.properties.get('gravitational_parameter')
    if mu is None:
        raise ValueError(
            'must pass in gravitational_parameter or '
            'set on Snap with set_gravitational_parameter'
        )

    radius = norm(pos, axis=1)
    Omega_k = np.sqrt(mu / radius ** 3)

    Stokes = t_s * Omega_k[:, np.newaxis]
    return Stokes


def temperature(
    snap: SnapLike, molecular_weight: float = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the gas temperature.

    Parameters
    ----------
    snap
        The Snap object.
    molecular_weight
        The gas molecular weight in gram / mole. E.g. 2.381 for
        molecular hydrogen.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The gas temperature on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        cs: Quantity = snap['sound_speed'][h > 0]
    else:
        cs = snap['sound_speed']

    gamma = snap.properties['adiabatic_index']

    molecular_weight = snap.properties.get('molecular_weight')
    if molecular_weight is None:
        logger.warning(
            'molecular_weight not set, assuming 2.381 for molecular hydrogen\n'
            'use set_molecular_weight to set on Snap'
        )
        molecular_weight = MOLECULAR_HYDROGEN_WEIGHT

    molecular_weight = molecular_weight * plonk_units('gram / mole')
    specific_gas_constant = (plonk_units.R / molecular_weight).to_base_units()

    return cs ** 2 / (gamma * specific_gas_constant)


def velocity_radial_cylindrical(
    snap: SnapLike, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the cylindrical radial velocity.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The radial velocity on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    x, y = pos[:, 0], pos[:, 1]
    vx, vy = vel[:, 0], vel[:, 1]

    return (x * vx + y * vy) / np.sqrt(x ** 2 + y ** 2)


def velocity_radial_spherical(
    snap: SnapLike, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the spherical radial velocity.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The radial velocity on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]
    vx, vy, vz = vel[:, 0], vel[:, 1], vel[:, 2]

    return (x * vx + y * vy + z * vz) / np.sqrt(x ** 2 + y ** 2 + z ** 2)



"""NumPy functions with Pint support."""

import numpy as np

from .._units import Quantity


def cross(x, y, **kwargs):
    """Cross product.

    Parameters
    ----------
    x, y
        The two arrays (N, 3) to take the cross product of. Can be
        ndarray or pint Quantity.
    **kwargs
        Keyword arguments to pass to np.cross.

    Returns
    -------
    ndarray
        The cross product of x and y.
    """
    if isinstance(x, Quantity):
        return np.cross(x.magnitude, y.magnitude, **kwargs) * x.units * y.units
    return np.cross(x, y, **kwargs)


def norm(x, **kwargs):
    """Norm of a vector.

    Parameters
    ----------
    x
        The arrays (N, 3) to take the norm of. Can be ndarray or pint
        Quantity.
    **kwargs
        Keyword arguments to pass to np.linalg.norm.

    Returns
    -------
    ndarray
        The norm of x.
    """
    if isinstance(x, Quantity):
        return np.linalg.norm(x.magnitude, **kwargs) * x.units
    return np.linalg.norm(x, **kwargs)


def average(x, weights, **kwargs):
    """Average.

    Parameters
    ----------
    x
        The array (N,) to take the norm of. Can be ndarray or pint
        Quantity.
    weights
        The weights for averaging.
    **kwargs
        Keyword arguments to pass to np.average.

    Returns
    -------
    ndarray
        The average of x.
    """
    if isinstance(x, Quantity):
        return np.average(x.magnitude, weights=weights.magnitude, **kwargs) * x.units
    return np.average(x, weights=weights, **kwargs)



"""
Plonk
=====

**Plonk** is a Python package for analyzing and visualizing smoothed
particle hydrodynamics simulation data.

Features
--------
Here are some of the features of Plonk:

- Read in Phantom HDF snapshot files.
- Read in global quantity and sink evolution files.
- Encapsulate entire simulation data as Simulation object.
- Access particle and sink arrays.
- Access simulation parameters and units.
- Compute extra quantities on particles.
- Visualize data using kernel interpolation.
- Generate profiles.

Classes
-------
Profile
    Represents a profile through the snapshot in Cartesian, cylindrical
    or spherical coordinates.
Sinks
    Represents sinks in a snapshot file.
Snap
    Represents a smoothed particle hydrodynamics snapshot file,
    containing particles, sinks, and file header information.
SubSnap
    Represents a subset of particles in a snapshot file.
Simulation
    Represents an entire smoothed particle hydrodynamics simulation.
    It contains a list of Snap objects, and time series data as pandas
    dataframes.

Subpackages
-----------
analysis
    Perform analysis on SPH snapshot data.
simulation
    Access multiple simulation files as a coherent object.
snap
    Read and accessing snapshot files.
utils
    Utility functions.
visualize
    Visualize of snapshot files.

Documentation
-------------
See https://plonk.readthedocs.io/ for documentation. The source code is
available at https://github.com/dmentipl/plonk.
"""

import importlib_metadata

from ._logging import logger_init as _logger_init
from ._units import add_units, array_units, units
from .analysis.profile import Profile, load_profile
from .simulation.evolution import load_ev
from .simulation.simulation import Simulation, load_sim
from .snap.readers import load_snap
from .snap.snap import Sinks, Snap, SnapLike, SubSnap
from .visualize.animation import animation, animation_particles, animation_profiles
from .visualize.interpolation import interpolate
from .visualize.simulation import visualize_sim
from .visualize.visualization import image, plot, vector

__version__ = importlib_metadata.version('plonk')

_logger_init(__version__)

add_units()

__all__ = [
    'Profile',
    'Simulation',
    'Sinks',
    'Snap',
    'SnapLike',
    'SubSnap',
    'add_units',
    'animation',
    'animation_particles',
    'animation_profiles',
    'array_units',
    'image',
    'interpolate',
    'load_ev',
    'load_profile',
    'load_sim',
    'load_snap',
    'plot',
    'units',
    'vector',
    'visualize_sim',
]



"""Simulation class for smoothed particle hydrodynamics simulations.

This class contains all information associated with a simulation. It
contains the snapshot as a list of Snap objects, and the global quantity
and sink quantity time series files as pandas DataFrames.
"""

from __future__ import annotations

from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Union

import numpy as np
from pandas import DataFrame

from .._logging import logger
from .._units import Quantity
from ..snap.readers import load_snap
from ..visualize.simulation import visualize_sim
from .evolution import evolution_units, load_ev

if TYPE_CHECKING:
    from ..snap.snap import Snap

_properties_vary_per_snap = ('time',)
_data_sources = ('Phantom',)


class Simulation:
    """Smoothed particle hydrodynamics simulation object.

    This class aggregates snapshot files, global quantity and sink time
    series data. Snapshot files contain a complete snapshot of the
    simulation at a particular time. Other files contain time series of
    global quantities on particles such as energy and momentum, and
    time series data for sink particles.

    Examples
    --------
    Reading simulation data into a Simulation object.

    >>> sim = plonk.load_sim('prefix', path_to_directory)

    Accessing the snapshots.

    >>> sim.snaps

    Accessing the properties.

    >>> sim.properties

    Accessing the global quantity and sink time series data.

    >>> sim.time_series['global']
    >>> sim.time_series['sinks']
    """

    def __init__(self):

        self.data_source: str
        self.prefix: str
        self.paths: Dict[str, Any]

        self._snaps: List[Snap] = None
        self._properties: Dict[str, Any] = None
        self._code_units: Dict[str, Any] = None
        self._time_series: Dict[str, Union[DataFrame, List[DataFrame]]] = None

        self._snap_file_extension = ''
        self._len = -1

    def load_sim(
        self,
        prefix: str,
        directory: Union[str, Path] = None,
        data_source: str = 'Phantom',
    ) -> Simulation:
        """Load Simulation.

        Parameters
        ----------
        prefix
            Simulation prefix, e.g. 'disc', if files are named like
            disc_00000.h5, disc01.ev, discSink0001N01.ev, etc.
        directory : optional
            Directory containing simulation snapshot files and auxiliary
            files. Default is None.
        data_source : optional
            The SPH code used to produce the simulation data. Default
            is 'Phantom'.
        """
        if data_source not in _data_sources:
            raise ValueError(f'Data source not available: try {_data_sources}')
        self.data_source = data_source

        if directory is None:
            directory = '.'

        logger.debug(f'Loading {data_source} simulation: {prefix} at {directory}')
        self.prefix = prefix
        self.paths = {
            'directory': Path(directory).expanduser().resolve(),
        }

        if not list(self.paths['directory'].glob(self.prefix + '*')):
            raise FileNotFoundError(f'No files with prefix: {prefix}')

        self._snap_file_extension = self._get_snap_file_extension()

        self.paths['snaps'] = self._get_snap_files()
        self.paths['time_series_global'] = self._get_global_ev_files()
        self.paths['time_series_sinks'] = self._get_sink_ev_files()

        return self

    @property
    def snaps(self) -> List[Snap]:
        """List of Snap objects associated with the simulation."""
        if self._snaps is None:
            self._generate_snap_objects()

        return self._snaps

    @property
    def properties(self) -> Dict[str, Any]:
        """Properties associated with the simulation."""
        if self._properties is None:
            self._generate_properties()

        return self._properties

    @property
    def code_units(self) -> Dict[str, Any]:
        """Units associated with the simulation."""
        if self._code_units is None:
            self._generate_units()

        return self._code_units

    @property
    def time_series(self) -> DataFrame:
        """Time series data."""
        if self._time_series is None:
            self._generate_time_series()

        return self._time_series

    def to_array(self, quantity: str, indices: List[int] = None) -> Quantity:
        """Generate an array of a quantity over all snapshots.

        Warning: this can be very memory intensive and slow.

        Parameters
        ----------
        quantity
            The quantity as a string, e.g. 'position'.
        indices
            You can select a subset of particles by indices
            corresponding to snap['id'].

        Returns
        -------
        An array with units.

        Examples
        --------
        Get the position of every particle during the whole simulation.

        >>> pos = sim.to_array(quantity='position')
        >>> pos.shape
            (31, 1100000, 3)
        """
        q = list()
        arr: Quantity = self.snaps[0][quantity]
        units = arr.units
        for snap in self.snaps:
            if indices is None:
                array: Quantity = snap[quantity]
            else:
                array = snap[quantity][indices]
            q.append(array.magnitude)
        q *= units
        return q

    def _generate_snap_objects(self):
        """Generate Snap objects."""
        snaps = list()
        fail = 0
        for snap in self.paths['snaps']:
            try:
                snaps.append(load_snap(snap))
            except (OSError, RuntimeError):
                fail += 1
        if fail > 0:
            logger.warning(f'Cannot read {fail} snap(s)')
        self._snaps = snaps

    def _generate_properties(self):
        """Generate sim.properties from snap.properties."""
        prop = copy(self.snaps[0].properties)
        for key in _properties_vary_per_snap:
            prop[key] = list()
        for snap in self.snaps:
            for key, val in snap.properties.items():
                if isinstance(prop[key], list):
                    prop[key].append(val)
                else:
                    if np.any(prop[key] != val):
                        prop[key] = '__inconsistent__'
        for key, val in prop.items():
            if isinstance(val, list):
                if isinstance(val[0], Quantity):
                    prop[key] = np.array([v.m for v in val]) * val[0].u
                else:
                    prop[key] = np.array(val)
        self._properties = prop

    def _generate_units(self):
        """Generate sim.code_units from snap.code_units."""
        u = copy(self.snaps[0].code_units)
        for snap in self.snaps:
            for key, val in snap.code_units.items():
                if u[key] != val:
                    u[key] = '__inconsistent__'
        self._code_units = u

    def _generate_time_series(self):
        """Generate time series data."""
        self._time_series = dict()
        if self.paths['time_series_global']:
            self._time_series['global'] = load_ev(self.paths['time_series_global'])
        if self.paths['time_series_sinks']:
            self._time_series['sinks'] = [
                load_ev(files) for files in self.paths['time_series_sinks']
            ]

    def _get_global_ev_files(self, glob: str = None) -> List[Path]:
        """Get global ev files."""
        if glob is None:
            # Phantom ev file name format
            glob = self.prefix + '[0-9][0-9].ev'

        return sorted(list(self.paths['directory'].glob(glob)))

    def _get_sink_ev_files(self, glob: str = None) -> List[List[Path]]:
        """Get sink ev files."""
        if glob is None:
            # Phantom ev file name format
            glob = self.prefix + 'Sink[0-9][0-9][0-9][0-9]N[0-9][0-9].ev'

        n = len(self.prefix) + len('Sink')
        n_sinks = len({p.name[n : n + 4] for p in self.paths['directory'].glob(glob)})

        sinks = list()
        for idx in range(1, n_sinks + 1):
            sinks.append(
                sorted(
                    self.paths['directory'].glob(
                        self.prefix + f'Sink{idx:04}N[0-9][0-9].ev'
                    )
                )
            )

        return sinks

    def set_units_on_time_series(self, config: Union[str, Path] = None):
        """Set physical units on time series data.

        Parameters
        ----------
        config : optional
            The path to a Plonk config.toml file.
        """
        units = evolution_units(sim=self, data_source=self.data_source, config=config)
        if 'global' in self.time_series:
            _apply_units_to_dataframe(self.time_series['global'], units)
        if 'sinks' in self.time_series:
            for ev in self.time_series['sinks']:
                _apply_units_to_dataframe(ev, units)

        return self

    def unset_units_on_time_series(self, config: Union[str, Path] = None):
        """Un-set physical units on time series data.

        Parameters
        ----------
        config : optional
            The path to a Plonk config.toml file.
        """
        units = evolution_units(sim=self, data_source=self.data_source, config=config)
        if 'global' in self.time_series:
            _un_apply_units_to_dataframe(self.time_series['global'], units)
        if 'sinks' in self.time_series:
            for ev in self.time_series['sinks']:
                _un_apply_units_to_dataframe(ev, units)

        return self

    def _get_snap_files(self, glob: str = None) -> List[Path]:
        """Get snapshot files."""
        if glob is None:
            # Phantom snapshot file name format
            glob = (
                self.prefix + '_[0-9][0-9][0-9][0-9][0-9].' + self._snap_file_extension
            )

        return sorted(list(self.paths['directory'].glob(glob)))

    def _get_snap_file_extension(self, glob: str = None):
        """Snapshot file extension.

        Determine snap file type from extension assuming file names
        follow a glob pattern.
        """
        if glob is None:
            # Phantom HDF5 snapshot file name format
            glob = self.prefix + '_[0-9][0-9][0-9][0-9][0-9].h5'

        file_types = {f.suffix for f in self.paths['directory'].glob(glob)}

        if len(file_types) > 1:
            raise ValueError(
                'Cannot determine simulation snapshot file type: '
                f'is it one of {file_types}?'
            )
        if len(file_types) == 0:
            raise ValueError(
                'Cannot determine snapshot file type: '
                'no files named like prefix_xxxxx.ext'
            )

        file_ext = file_types.pop()[1:]

        if file_ext not in ('h5',):
            raise ValueError('File extension not available; must be ".h5"')

        return file_ext

    def __len__(self):
        """Length as number of snaps."""
        if self._len == -1:
            self._len = len(self.snaps)
        return self._len

    def __repr__(self):
        """Dunder repr method."""
        return self.__str__()

    def __str__(self):
        """Dunder str method."""
        return (
            f'<plonk.Simulation: "{self.prefix}", '
            f'directory="{self.paths["directory"].name}">'
        )

    visualize = visualize_sim


def load_sim(
    prefix: str, directory: Union[str, Path] = None, data_source: str = 'Phantom',
) -> Simulation:
    """Load Simulation.

    Parameters
    ----------
    prefix
        Simulation prefix, e.g. 'disc', if files are named like
        disc_00000.h5, disc01.ev, discSink0001N01.ev, etc.
    directory : optional
        Directory containing simulation snapshot files and auxiliary
        files. Default is None.
    data_source : optional
        The SPH code used to produce the simulation data. Default
        is 'Phantom'.
    """
    return (
        Simulation()
        .load_sim(prefix=prefix, directory=directory, data_source=data_source)
        .set_units_on_time_series()
    )


def _apply_units_to_dataframe(dataframe, units):
    keys = list()
    for key, val in units.items():
        if key in dataframe:
            keys.append(key)
            dataframe[key] = (dataframe[key].to_numpy() * units[key]).magnitude
    mapper = {key: f'{key} [{units[key].units:~}]' for key in keys}
    dataframe.rename(columns=mapper, inplace=True)
    return dataframe


def _un_apply_units_to_dataframe(dataframe, units):
    keys = list()
    for key, val in units.items():
        key_unit = f'{key} [{units[key].units:~}]'
        if key_unit in dataframe:
            keys.append(key)
            dataframe[key_unit] = (
                dataframe[key_unit].to_numpy() / units[key]
            ).magnitude
    mapper = {f'{key} [{units[key].units:~}]': key for key in keys}
    dataframe.rename(columns=mapper, inplace=True)
    return dataframe



"""Testing visualization.interpolation."""

import numpy as np

from plonk.visualize.interpolation import scalar_interpolation, vector_interpolation

from .stubdata.interpolation_arrays import (
    scalar_projection,
    scalar_slice,
    vector_projection,
    vector_slice,
)

N = 10

XX = np.ones(N)
YY = np.ones(N)
ZZ = np.ones(N)
HH = np.ones(N)
WW = np.ones(N)
MM = np.ones(N)

S_DATA = np.ones(N)
X_DATA = np.ones(N)
Y_DATA = np.ones(N)

EXTENT = (0, 1, 0, 1)
PIX = (10, 10)
ZSLICE = 0.5
HFACT = 1.0


def test_scalar_interpolation_projection():
    """Test projection interpolation."""
    im = scalar_interpolation(
        quantity=S_DATA,
        x_coordinate=XX,
        y_coordinate=YY,
        extent=EXTENT,
        smoothing_length=HH,
        particle_mass=MM,
        hfact=HFACT,
        number_of_pixels=PIX,
    )

    np.testing.assert_allclose(im, scalar_projection, rtol=1e-5)


def test_scalar_interpolation_slice():
    """Test cross section interpolation."""
    im = scalar_interpolation(
        quantity=S_DATA,
        x_coordinate=XX,
        y_coordinate=YY,
        dist_from_slice=ZZ * ZSLICE,
        extent=EXTENT,
        smoothing_length=HH,
        particle_mass=MM,
        hfact=HFACT,
        number_of_pixels=PIX,
    )

    np.testing.assert_allclose(im, scalar_slice, rtol=1e-5)


def test_vector_interpolation_projection():
    """Test projection interpolation."""
    vec = vector_interpolation(
        quantity_x=X_DATA,
        quantity_y=Y_DATA,
        x_coordinate=XX,
        y_coordinate=YY,
        extent=EXTENT,
        smoothing_length=HH,
        particle_mass=MM,
        hfact=HFACT,
        number_of_pixels=PIX,
    )

    np.testing.assert_allclose(vec, vector_projection, rtol=1e-5)


def test_vector_interpolation_slice():
    """Test cross section interpolation."""
    vec = vector_interpolation(
        quantity_x=X_DATA,
        quantity_y=Y_DATA,
        x_coordinate=XX,
        y_coordinate=YY,
        dist_from_slice=ZZ * ZSLICE,
        extent=EXTENT,
        smoothing_length=HH,
        particle_mass=MM,
        hfact=HFACT,
        number_of_pixels=PIX,
    )

    np.testing.assert_allclose(vec, vector_slice, rtol=1e-5)



"""Config."""

from pathlib import Path
from typing import Any, MutableMapping, Union

import toml

CONFIG_FILE = Path(__file__).parent / 'config.toml'


def load_config(filename: Union[str, Path] = None) -> MutableMapping[str, Any]:
    """Load config file.

    Parameters
    ----------
    filename
        The name of the config file (must be TOML) as a string or
        pathlib.Path.

    Returns
    -------
    Dict
        The config as a nested dictionary.
    """
    if filename is None:
        filename = CONFIG_FILE
    return toml.load(filename)


def write_config(filename: Union[str, Path]):
    """Write a default config file.

    Parameters
    ----------
    filename
        The name of the config file (must be TOML) as a string or
        pathlib.Path.
    """
    config = load_config()
    with open(filename, mode='w') as f:
        toml.dump(config, f)



"""Visualize SPH data.

The Plonk implementation for visualizing smoothed particle hydrodynamics
simulations using kernel density estimation based interpolation.
"""

from .animation import animation, animation_particles, animation_profiles
from .functions import get_extent_from_percentile, plot_smoothing_length
from .interpolation import interpolate
from .simulation import visualize_sim
from .visualization import image, plot, vector

__all__ = [
    'animation',
    'animation_particles',
    'animation_profiles',
    'get_extent_from_percentile',
    'image',
    'plot_smoothing_length',
    'interpolate',
    'plot',
    'vector',
    'visualize_sim',
]



"""Testing Simulation."""

from pathlib import Path

import numpy as np
import pytest

import plonk


def test_init_simulation():
    """Testing initialising simulation."""
    dir_path = Path(__file__).parent / 'stubdata'
    plonk.load_sim(prefix='phantom', directory=dir_path)
    with pytest.raises(ValueError):
        plonk.load_sim(
            prefix='phantom', directory=dir_path, data_source='not_available'
        )
    with pytest.raises(FileNotFoundError):
        plonk.load_sim(prefix='does_not_exist', directory=dir_path)


def test_sim_data():
    """Testing data in simulation."""
    dir_path = Path(__file__).parent / 'stubdata'
    sim = plonk.load_sim(prefix='phantom', directory=dir_path)

    snaps = sim.snaps
    assert len(snaps) == 1
    assert len(snaps[0]) == 2000

    properties = {
        'adiabatic_index': 1.0,
        'dust_method': 'dust as separate sets of particles',
        'equation_of_state': 'locally isothermal disc',
        'grain_density': [3000.0] * plonk.units('kg/m^3'),
        'grain_size': [0.01] * plonk.units('m'),
        'smoothing_length_factor': 1.0,
        'time': [0.0] * plonk.units('s'),
    }

    for key, val in sim.properties.items():
        if isinstance(val, plonk._units.Quantity):
            assert np.allclose(val.m, properties[key].m)
        else:
            assert sim.properties[key] == properties[key]

    assert sim.paths['time_series_global'][0].name == 'phantom01.ev'


def test_simulation_visualization():
    """Test simulation visualization."""
    dir_path = Path(__file__).parent / 'stubdata'
    sim = plonk.load_sim(prefix='phantom', directory=dir_path)

    viz = sim.visualize(kind='particle', x='x', y='y')
    viz.next()
    viz.prev()


def test_to_array():
    """Testing to_array method."""
    dir_path = Path(__file__).parent / 'stubdata'
    sim = plonk.load_sim(prefix='phantom', directory=dir_path)
    sim.to_array(quantity='density', indices=[0, 1, 2])


def test_set_units_time_series():
    """Test set/unset units time series."""
    dir_path = Path(__file__).parent / 'stubdata'
    sim = plonk.load_sim(prefix='phantom', directory=dir_path)
    sim.unset_units_on_time_series()
    sim.set_units_on_time_series()



"""Calculate extra quantities on the sinks."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numba
import numpy as np

from .._units import Quantity
from .._units import units as plonk_units

if TYPE_CHECKING:
    from ..snap.snap import Sinks

G = (1 * plonk_units.newtonian_constant_of_gravitation).to_base_units()


def kinetic_energy(sinks: Sinks) -> Quantity:
    """Calculate the total kinetic energy.

    Parameters
    ----------
    sinks
        The Sinks object.

    Returns
    -------
    Quantity
    """
    m, v = sinks['mass'], sinks['velocity']
    return np.sum(1 / 2 * m * _norm(v) ** 2)


def gravitational_potential_energy(sinks: Sinks) -> Quantity:
    """Calculate the total gravitational potential energy.

    Parameters
    ----------
    sinks
        The Sinks object.

    Returns
    -------
    Quantity
    """
    position = sinks['position']
    mass = sinks['mass']
    potential = _potential(position=position.magnitude, mass=mass.magnitude)
    return (potential * G * mass.units ** 2 / position.units).to_base_units()


def specific_orbital_energy(sinks: Sinks) -> Quantity:
    """Calculate the specific orbital energy for two bodies.

    Parameters
    ----------
    sinks
        The Sinks object. Must have length 2.

    Returns
    -------
    Quantity
    """
    if len(sinks) != 2:
        raise ValueError('sinks must have length 2')
    m1, m2 = sinks['mass']
    mu = m1 * m2 / (m1 + m2)
    ke = kinetic_energy(sinks=sinks)
    pe = gravitational_potential_energy(sinks=sinks)
    return (ke + pe) / mu


def specific_angular_momentum(sinks: Sinks) -> Quantity:
    """Calculate the specific orbital energy for two bodies.

    Parameters
    ----------
    sinks
        The Sinks object. Must have length 2.

    Returns
    -------
    Quantity
        The specific angular momentum.
    """
    if len(sinks) != 2:
        raise ValueError('sinks must have length 2')
    r = sinks['position'][1] - sinks['position'][0]
    v = sinks['velocity'][1] - sinks['velocity'][0]
    return np.cross(r, v)


def eccentricity(sinks: Sinks) -> Quantity:
    """Calculate the eccentricity.

    Parameters
    ----------
    sinks
        The Sinks object. Must have length 2.

    Returns
    -------
    Quantity
    """
    if len(sinks) != 2:
        raise ValueError('sinks must have length 2')
    mu = G * np.sum(sinks['mass'])
    eps = specific_orbital_energy(sinks)
    h = specific_angular_momentum(sinks)
    h_mag = _norm(h)
    return np.sqrt(1 + 2 * eps * h_mag ** 2 / mu ** 2)


def semi_major_axis(sinks: Sinks) -> Quantity:
    """Calculate the semi-major axis for two bodies.

    Parameters
    ----------
    sinks
        The Sinks object. Must have length 2.

    Returns
    -------
    Quantity
    """
    if len(sinks) != 2:
        raise ValueError('sinks must have length 2')
    mu = G * np.sum(sinks['mass'])
    h = specific_angular_momentum(sinks)
    h_mag = _norm(h)
    e = eccentricity(sinks)
    return h_mag ** 2 / (mu * (1 - e ** 2))


def inclination(sinks: Sinks) -> Quantity:
    """Calculate the inclination for two bodies.

    Parameters
    ----------
    sinks
        The Sinks object. Must have length 2.

    Returns
    -------
    Quantity
    """
    if len(sinks) != 2:
        raise ValueError('sinks must have length 2')
    h = specific_angular_momentum(sinks)
    h_z = h[..., 2]
    h_mag = _norm(h)
    return np.arccos(h_z / h_mag)


def orbital_period(sinks: Sinks) -> Quantity:
    """Calculate the orbital period for two bodies.

    Parameters
    ----------
    sinks
        The Sinks object. Must have length 2.

    Returns
    -------
    Quantity
    """
    if len(sinks) != 2:
        raise ValueError('sinks must have length 2')
    a = semi_major_axis(sinks)
    mu = G * np.sum(sinks['mass'])
    return 2 * np.pi * np.sqrt(a ** 3 / mu)


def mean_motion(sinks: Sinks) -> Quantity:
    """Calculate the mean motion for two bodies.

    Parameters
    ----------
    sinks
        The Sinks object. Must have length 2.

    Returns
    -------
    Quantity
    """
    if len(sinks) != 2:
        raise ValueError('sinks must have length 2')
    P = orbital_period(sinks)
    return 2 * np.pi / P


def Roche_sphere(sinks: Sinks) -> Quantity:
    """Calculate an estimate of the Roche sphere for two bodies.

    The Roche sphere radius is calculated around the first of the two
    sinks. Uses the formula from Eggleton (1983) ApJ 268, 368-369.

    Parameters
    ----------
    sinks
        The Sinks object. Must have length 2.

    Returns
    -------
    Quantity
    """
    if len(sinks) != 2:
        raise ValueError('sinks must have length 2')
    separation = _norm(sinks['position'][1] - sinks['position'][0])
    q = sinks['mass'][0] / sinks['mass'][1]
    return (
        separation
        * 0.49
        * q ** (2 / 3)
        / (0.6 * q ** (2 / 3) + np.log(1.0 + q ** (1 / 3)))
    )


def Hill_radius(primary: Sinks, secondary: Sinks) -> Quantity:
    """Calculate the Hill radius.

    Parameters
    ----------
    primary
        The primary, i.e. heavy object, as a Sinks object. It must have
        one sink.
    secondary
        The secondary, i.e. light objects, as a Sinks object. It can
        have more than one sink, e.g. when calculating the Hill radiius
        for multiple planets orbiting a star.

    Returns
    -------
    Quantity
    """
    if len(primary) != 1:
        raise ValueError('primary must have length 1')
    Hill = [_Hill_radius(primary + sink) for sink in secondary]
    return [H.magnitude for H in Hill] * Hill[0].units


def _Hill_radius(sinks):
    if len(sinks) != 2:
        raise ValueError('sinks must have length 2')
    M, m = sinks['mass']
    q = m / M if m / M < 1 else M / m
    a = semi_major_axis(sinks)
    e = eccentricity(sinks)
    return a * (1 - e) * (q / 3) ** (1 / 3)


def _norm(x):
    return np.sqrt(x[..., 0] ** 2 + x[..., 1] ** 2 + x[..., 2] ** 2)


@numba.njit
def _potential(position, mass):
    """Get gravitational potential on particles.

    Parameters
    ----------
    position
        Particle positions.
    mass
        Particle masses.

    Returns
    -------
    potential
        The total gravitational potential.
    """
    number_of_particles = len(mass)
    potential = 0.0

    # Loop over all particles...
    for i in range(number_of_particles):
        # ...and all neighbours
        phi = 0.0
        for j in range(number_of_particles):
            # Ignore self
            if j == i:
                continue
            dx = position[i, :] - position[j, :]
            r = np.sqrt(dx[0] ** 2 + dx[1] ** 2 + dx[2] ** 2)
            phi += -mass[j] / r
        potential += mass[i] * phi

    # Account for double counting
    potential /= 2

    return potential



"""Units."""

from pathlib import Path
from typing import Any, Dict, Union

import pint

from ._config import load_config

units = pint.UnitRegistry()
Quantity = units.Quantity


def add_units(config: Union[str, Path] = None):
    """Add units to the unit registry from a config file.

    Parameters
    ----------
    config : optional
        The path to a Plonk config.toml file. If None, use the default
        file.
    """
    conf = load_config(filename=config)
    for unit, definition in conf['units']['definitions'].items():
        units.define(f'{unit} = {definition}')


def array_units(config: Union[str, Path] = None) -> Dict[str, str]:
    """Return a dictionary of arrays with unit strings.

    Like the following:

        `{'density': 'kg / m ^ 3', ... 'position': 'm', ... }`

    Parameters
    ----------
    config : optional
        The path to a Plonk config.toml file. If None, use the default
        file.

    Returns
    -------
    Dict
    """
    conf = load_config(filename=config)
    d = dict()
    for key, val in conf['arrays']['dimensions'].items():
        dim = _convert_dim_string(val)
        if dim == {'[angle]': 1.0}:
            d[key] = "radian"
        else:
            for key1, val1 in conf['units']['defaults'].items():
                if _dimensionality_comparison(dict(units(val1).dimensionality), dim):
                    d[key] = val1
                    break
    return d


def array_quantities(config: Union[str, Path] = None) -> Dict[str, Any]:
    """Array dimensions dictionary.

    Like the following:

        `{
            'density': {'[mass]': 1, '[length]': -3},
            ...
            'position': {'[length]': 1},
            ...
        }`

    Parameters
    ----------
    config : optional
        The path to a Plonk config.toml file. If None, use the default
        file.

    Returns
    -------
    Dict
    """
    conf = load_config(filename=config)
    arrays = conf['arrays']['dimensions']
    dim = dict()
    for key, val in arrays.items():
        dim[key] = _convert_dim_string(val)
    return dim


def generate_array_code_units(code_units: Dict[str, Any]) -> Dict[str, Any]:
    """Generate array code units dictionary.

    Like the following:

        `{
            'density': 1.0 <Unit('kilogram / meter ** 3')>,
            ...
            'position': 149600000000.0 <Unit('meter')>,
            ...
        }`

    Parameters
    ----------
    code_units

    Returns
    -------
    Dict
        A dictionary of units as Pint quantities.
    """
    _units = dict()
    _array_quantities = array_quantities()
    for arr, unit in _array_quantities.items():
        _units[arr] = _get_code_unit(unit, code_units)
    return _units


def _convert_dim_string(string):
    if string == '':
        return {}
    list_dims = [dim.strip() for dim in string.split(',')]
    keys = list()
    vals = list()
    for dim in list_dims:
        key, val = dim.split(': ')
        val = float(val)
        keys.append(key)
        vals.append(val)
    return {'[' + key + ']': val for key, val in zip(keys, vals)}


def _get_code_unit(dim_dict, code_units):
    unit = 1.0 * units['dimensionless']
    for d in dim_dict:
        if d == '[angle]':
            unit *= units['radian']
        else:
            unit *= code_units[d[1:-1]] ** dim_dict[d]
    return unit


def _dimensionality_comparison(dim1, dim2):
    _dim1 = {key: float(val) for key, val in dim1.items()}
    _dim2 = {key: float(val) for key, val in dim2.items()}
    return _dim1 == _dim2



"""Test animation functions."""

from pathlib import Path

import plonk


def test_animation():
    """Test animation of images."""
    dir_path = Path(__file__).parent / 'stubdata'
    sim = plonk.load_sim(prefix='phantom', directory=dir_path)

    snaps = [sim.snaps[0], sim.snaps[0], sim.snaps[0]]
    filename = Path('animation.mp4')
    plonk.animation(
        filename=filename,
        snaps=snaps,
        quantity='density',
        units={'position': 'au', 'density': 'g/cm^3'},
        adaptive_colorbar=False,
        number_of_pixels=(32, 32),
    )
    filename.unlink()


def test_animation_profiles():
    """Test animation of profiles."""
    dir_path = Path(__file__).parent / 'stubdata'
    sim = plonk.load_sim(prefix='phantom', directory=dir_path)

    snaps = [sim.snaps[0], sim.snaps[0], sim.snaps[0]]
    profiles = [plonk.load_profile(snap) for snap in snaps]

    filename = Path('animation.mp4')
    plonk.animation_profiles(
        filename=filename,
        profiles=profiles,
        x='radius',
        y='surface_density',
        units={'position': 'au', 'surface_density': 'g/cm^2'},
    )
    filename.unlink()


def test_animation_particles():
    """Test animation of particle plots."""
    dir_path = Path(__file__).parent / 'stubdata'
    sim = plonk.load_sim(prefix='phantom', directory=dir_path)

    snaps = [sim.snaps[0], sim.snaps[0], sim.snaps[0]]
    filename = Path('animation.mp4')
    plonk.animation_particles(
        filename=filename,
        snaps=snaps,
        x='x',
        y='density',
        units={'position': 'au', 'density': 'g/cm^3'},
        adaptive_limits=False,
    )
    filename.unlink()



"""Test analysis functions."""

from pathlib import Path

import plonk
from plonk import analysis

TEST_FILE = Path(__file__).parent / 'stubdata/phantom_00000.h5'
AU = plonk.units('au')


def test_particles():
    """Test particles analysis functions."""
    snap = plonk.load_snap(TEST_FILE)

    snap.set_molecular_weight(2.381)
    snap.set_gravitational_parameter(0)
    mu = snap.properties['gravitational_parameter']

    _test_particles(snap=snap, ignore=False, mu=mu)
    _test_particles(snap=snap, ignore=True, mu=mu)

    snap.close_file()


def _test_particles(snap, ignore, mu):

    analysis.particles.angular_momentum(snap=snap, ignore_accreted=ignore)
    analysis.particles.angular_velocity(snap=snap, ignore_accreted=ignore)
    analysis.particles.azimuthal_angle(snap=snap, ignore_accreted=ignore)
    analysis.particles.dust_density(snap=snap, ignore_accreted=ignore)
    analysis.particles.dust_fraction(snap=snap, ignore_accreted=ignore)
    analysis.particles.dust_mass(snap=snap, ignore_accreted=ignore)
    analysis.particles.gas_density(snap=snap, ignore_accreted=ignore)
    analysis.particles.gas_fraction(snap=snap, ignore_accreted=ignore)
    analysis.particles.gas_mass(snap=snap, ignore_accreted=ignore)
    analysis.particles.eccentricity(
        snap=snap, gravitational_parameter=mu, ignore_accreted=ignore
    )
    analysis.particles.inclination(snap=snap, ignore_accreted=ignore)
    analysis.particles.keplerian_frequency(
        snap=snap, gravitational_parameter=mu, ignore_accreted=ignore
    )
    analysis.particles.kinetic_energy(snap=snap, ignore_accreted=ignore)
    analysis.particles.momentum(snap=snap, ignore_accreted=ignore)
    analysis.particles.polar_angle(snap=snap, ignore_accreted=ignore)
    analysis.particles.radius_cylindrical(snap=snap, ignore_accreted=ignore)
    analysis.particles.radius_spherical(snap=snap, ignore_accreted=ignore)
    analysis.particles.semi_major_axis(
        snap=snap, gravitational_parameter=mu, ignore_accreted=ignore
    )
    analysis.particles.specific_angular_momentum(snap=snap, ignore_accreted=ignore)
    analysis.particles.specific_kinetic_energy(snap=snap, ignore_accreted=ignore)
    analysis.particles.stokes_number(
        snap=snap, gravitational_parameter=mu, ignore_accreted=ignore
    )
    analysis.particles.temperature(snap=snap, ignore_accreted=ignore)


def test_total():
    """Test total analysis functions."""
    snap = plonk.load_snap(TEST_FILE)

    analysis.total.accreted_mass(snap=snap)
    analysis.total.angular_momentum(snap=snap)
    analysis.total.center_of_mass(snap=snap)
    analysis.total.inclination(snap=snap)
    analysis.total.kinetic_energy(snap=snap)
    analysis.total.mass(snap=snap)
    analysis.total.momentum(snap=snap)
    analysis.total.position_angle(snap=snap)
    analysis.total.specific_angular_momentum(snap=snap)
    analysis.total.specific_kinetic_energy(snap=snap)

    snap.close_file()


def test_filters():
    """Test particle filter functions."""
    snap = plonk.load_snap(TEST_FILE)

    xwidth, ywidth, zwidth = 10 * AU, 10 * AU, 10 * AU
    height = 10 * AU
    radius = 100 * AU
    radius_min = 10 * AU
    radius_max = 20 * AU

    analysis.filters.annulus(
        snap=snap, radius_min=radius_min, radius_max=radius_max, height=height
    )
    analysis.filters.box(snap=snap, xwidth=xwidth, ywidth=ywidth, zwidth=zwidth)
    analysis.filters.cylinder(snap=snap, radius=radius, height=height)
    analysis.filters.shell(snap=snap, radius_min=radius_min, radius_max=radius_max)
    analysis.filters.sphere(snap=snap, radius=radius)

    snap.close_file()


def test_discs():
    """Test discs analysis functions."""
    snap = plonk.load_snap(TEST_FILE)

    analysis.discs.normal(snap)
    analysis.discs.rotate_edge_on(snap)
    analysis.discs.rotate_face_on(snap)

    snap.close_file()



"""Filter particles to produce SubSnaps."""

from .._units import Quantity
from .._units import units as plonk_units
from ..snap.snap import SnapLike, SubSnap
from .particles import radius_cylindrical, radius_spherical

CENTER = (0, 0, 0) * plonk_units.au


def box(
    snap: SnapLike,
    xwidth: Quantity,
    ywidth: Quantity,
    zwidth: Quantity,
    center: Quantity = CENTER,
) -> SubSnap:
    """Particles within a box.

    Parameters
    ----------
    snap
        The Snap object.
    xwidth
        The x-width of the box.
    ywidth
        The y-width of the box.
    zwidth
        The z-width of the box.
    center : optional
        The center of the box as a Quantity like (x, y, z) * au.
        Default is (0, 0, 0).

    Returns
    -------
    SubSnap
        The SubSnap with particles in the box.
    """
    dx, dy, dz = xwidth / 2, ywidth / 2, zwidth / 2
    mask = (
        (snap['position_x'] > center[0] - dx)
        & (snap['position_x'] < center[0] + dx)
        & (snap['position_y'] > center[1] - dy)
        & (snap['position_y'] < center[1] + dy)
        & (snap['position_z'] > center[2] - dz)
        & (snap['position_z'] < center[2] + dz)
    )
    return snap[mask]


def cylinder(
    snap: SnapLike, radius: Quantity, height: Quantity, center: Quantity = CENTER
) -> SubSnap:
    """Particles within a cylinder.

    Parameters
    ----------
    snap
        The Snap object.
    radius
        The radius of the cylinder.
    height
        The height of the cylinder.
    center : optional
        The center of the cylinder as a Quantity like (x, y, z) * au.
        Default is (0, 0, 0).

    Returns
    -------
    SubSnap
        The SubSnap with particles in the cylinder.
    """
    dh = height / 2
    R = radius_cylindrical(snap=snap, origin=center)
    mask = (
        (R < radius)
        & (snap['position_z'] < center[2] + dh)
        & (snap['position_z'] > center[2] - dh)
    )
    return snap[mask]


def annulus(
    snap: SnapLike,
    radius_min: Quantity,
    radius_max: Quantity,
    height: Quantity,
    center: Quantity = CENTER,
) -> SubSnap:
    """Particles within an annulus.

    Parameters
    ----------
    snap
        The Snap object.
    radius_min
        The inner radius of the annulus.
    radius_max
        The outer radius of the annulus.
    height
        The height of the annulus.
    center : optional
        The center of the annulus as a Quantity like (x, y, z) * au.
        Default is (0, 0, 0).

    Returns
    -------
    SubSnap
        The SubSnap with particles in the annulus.
    """
    dh = height / 2
    R = radius_cylindrical(snap=snap, origin=center)
    mask = (
        (R > radius_min)
        & (R < radius_max)
        & (snap['position_z'] < center[2] + dh)
        & (snap['position_z'] > center[2] - dh)
    )
    return snap[mask]


def sphere(snap: SnapLike, radius: Quantity, center: Quantity = CENTER) -> SubSnap:
    """Particles within a sphere.

    Parameters
    ----------
    snap
        The Snap object.
    radius
        The radius of the sphere.
    center : optional
        The center of the sphere as a Quantity like (x, y, z) * au.
        Default is (0, 0, 0).

    Returns
    -------
    SubSnap
        The SubSnap with particles in the sphere.
    """
    R = radius_spherical(snap=snap, origin=center)
    mask = R < radius
    return snap[mask]


def shell(
    snap: SnapLike,
    radius_min: Quantity,
    radius_max: Quantity,
    center: Quantity = CENTER,
) -> SubSnap:
    """Particles within a spherical shell.

    Parameters
    ----------
    snap
        The Snap object.
    radius_min
        The inner radius of the shell.
    radius_max
        The outer radius of the shell.
    center : optional
        The center of the shell as a Quantity like (x, y, z) * au.
        Default is (0, 0, 0).

    Returns
    -------
    SubSnap
        The SubSnap with particles in the shell.
    """
    R = radius_spherical(snap=snap, origin=center)
    mask = (R > radius_min) & (R < radius_max)
    return snap[mask]



"""Animations of visualizations."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Union

import matplotlib.pyplot as plt
from matplotlib import animation as _animation

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None

from .._logging import logger
from .._units import units as plonk_units
from . import visualization as viz

if TYPE_CHECKING:
    from ..analysis.profile import Profile
    from ..snap.snap import SnapLike

_interp_kwargs = ('number_of_pixels', 'density_weighted')


def animation(
    *,
    filename: Union[str, Path],
    snaps: List[SnapLike],
    quantity: str,
    fig: Any = None,
    adaptive_colorbar: bool = True,
    text: List[str] = None,
    text_kwargs: Dict[str, Any] = {},
    func_animation_kwargs: Dict[str, Any] = {},
    save_kwargs: Dict[str, Any] = {},
    **kwargs,
):
    """Generate an animation of images.

    Parameters
    ----------
    filename
        The file name to save the animation to.
    snaps
        A list of Snap objects to animate.
    quantity
        The quantity to visualize. Must be a string to pass to Snap.
    fig : optional
        A matplotlib Figure object to animate. If None, generate a new
        Figure.
    adaptive_colorbar : optional
        If True, adapt colorbar range during animation. If False, the
        colorbar range is fixed by the initial image. Default is True.
    text : optional
        List of strings to display per snap.
    text_kwargs : optional
        Keyword arguments to pass to matplotlib text.
    func_animation_kwargs : optional
        Keyword arguments to pass to matplotlib FuncAnimation.
    save_kwargs : optional
        Keyword arguments to pass to matplotlib Animation.save.
    **kwargs
        Arguments to pass to visualize.image.

    Returns
    -------
    anim
        The matplotlib FuncAnimation object.

    Examples
    --------
    Make an animation of projected density.

    >>> plonk.animation(
    ...     filename='animation.mp4',
    ...     snaps=snaps,
    ...     quantity='density',
    ...     units={'position': 'au', 'density': 'g/cm^3'},
    ...     save_kwargs={'fps': 10, 'dpi': 300},
    ... )
    """
    filepath = Path(filename)
    if filepath.suffix != '.mp4':
        raise ValueError('filename should end in ".mp4"')

    if fig is None:
        ax = snaps[0].image(quantity=quantity, **kwargs)
        fig = ax.figure
        image = ax.images[0]
    else:
        ax = fig.axes[0]
        image = ax.images[0]

    if text is not None:
        _text = ax.text(
            0.9, 0.9, text[0], ha='right', transform=ax.transAxes, **text_kwargs
        )

    if tqdm is not None:
        pbar = tqdm(total=len(snaps))
    else:
        logger.info(
            'progress bar not available\ntry pip install tqdm --or-- conda install tqdm'
        )

    x = kwargs.get('x', 'x')
    y = kwargs.get('y', 'y')
    interp = kwargs.get('interp', 'projection')
    slice_normal = kwargs.get('slice_normal')
    slice_offset = kwargs.get('slice_offset')
    extent = kwargs.get('extent')
    units = kwargs.get('units')
    interp_kwargs = {k: v for k, v in kwargs.items() if k in _interp_kwargs}

    def animate(idx):
        if tqdm is not None:
            pbar.update(n=1)
        snap = snaps[idx]
        interp_data, _extent, _units = viz._interpolated_data(
            snap=snap,
            quantity=quantity,
            x=x,
            y=y,
            interp=interp,
            slice_normal=slice_normal,
            slice_offset=slice_offset,
            extent=extent,
            units=units,
            **interp_kwargs,
        )
        image.set_data(interp_data)
        image.set_extent(_extent)
        image.axes.set_xlim(_extent[:2])
        image.axes.set_ylim(_extent[2:])
        if adaptive_colorbar:
            vmin = kwargs.get('vmin', interp_data.min())
            vmax = kwargs.get('vmax', interp_data.max())
            image.set_clim(vmin, vmax)
        if text is not None:
            _text.set_text(text[idx])
        return [image]

    anim = _animation.FuncAnimation(
        fig, animate, frames=len(snaps), **func_animation_kwargs
    )
    anim.save(filepath, extra_args=['-vcodec', 'libx264'], **save_kwargs)
    plt.close()
    if tqdm is not None:
        pbar.close()

    return anim


def animation_profiles(
    *,
    filename: Union[str, Path],
    profiles: List[Profile],
    x: str,
    y: Union[str, List[str]],
    fig: Any = None,
    adaptive_limits: bool = True,
    text: List[str] = None,
    text_kwargs: Dict[str, Any] = {},
    func_animation_kwargs: Dict[str, Any] = {},
    save_kwargs: Dict[str, Any] = {},
    **kwargs,
):
    """Generate an animation of a profile.

    Parameters
    ----------
    filename
        The file name to save the animation to.
    profiles
        A list of Profile objects to animate.
    x
        The quantity for the x-axis. Must be a string to pass to Snap.
    y
        The quantity for the y-axis, or list of quantities. Must be a
        string (or list of strings) to pass to Snap.
    fig : optional
        A matplotlib Figure object to animate. If None, generate a new
        Figure.
    adaptive_limits : optional
        If True, adapt plot limits during animation. If False, the plot
        limits are fixed by the initial plot. Default is True.
    text : optional
        List of strings to display per profile plot.
    text_kwargs : optional
        Keyword arguments to pass to matplotlib text.
    func_animation_kwargs : optional
        Keyword arguments to pass to matplotlib FuncAnimation.
    save_kwargs : optional
        Keyword arguments to pass to matplotlib Animation.save.
    **kwargs
        Arguments to pass to Profile.plot function.

    Returns
    -------
    anim
        The matplotlib FuncAnimation object.

    Examples
    --------
    Make an animation of radius vs surface density.

    >>> plonk.animation_profiles(
    ...     filename='animation.mp4',
    ...     profiles=profiles,
    ...     x='radius',
    ...     y='surface_density',
    ...     units={'position': 'au', 'surface_density': 'g/cm^2'},
    ...     adaptive_limits=False,
    ...     save_kwargs={'fps': 10, 'dpi': 300},
    ... )
    """
    filepath = Path(filename)
    if filepath.suffix != '.mp4':
        raise ValueError('filename should end in ".mp4"')

    if isinstance(y, str):
        ys = [y]
    elif isinstance(y, (tuple, list)):
        ys = y
    else:
        raise ValueError('y must be a string or list of strings')

    prof = profiles[0]
    if fig is None:
        ax = prof.plot(x=x, y=ys, **kwargs)
        fig = ax.figure
        lines = ax.lines
    else:
        ax = fig.axes[0]
        lines = [line for line in ax.lines]

    if text is not None:
        _text = ax.text(
            0.9, 0.9, text[0], ha='right', transform=ax.transAxes, **text_kwargs
        )

    if tqdm is not None:
        pbar = tqdm(total=len(profiles))
    else:
        logger.info(
            'progress bar not available\ntry pip install tqdm --or-- conda install tqdm'
        )

    units = kwargs.get('units')
    if units is not None:
        xunit = plonk_units(units.get(prof.base_array_name(x), str(prof[x].units)))
        yunits = [
            plonk_units(units.get(prof.base_array_name(y), str(prof[y].units)))
            for y in ys
        ]
    else:
        xunit = prof[x].units
        yunits = [prof[y].units for y in ys]

    def animate(idx):
        if tqdm is not None:
            pbar.update(n=1)
        profile = profiles[idx]
        xlim, ylim = (0, 0), (0, 0)
        _x = (profile[x] / xunit).to_base_units().magnitude
        for line, y, yunit in zip(lines, ys, yunits):
            _y = (profile[y] / yunit).to_base_units().magnitude
            line.set_data(_x, _y)
            if adaptive_limits:
                ax = line.axes
                xlim, ylim = _get_range(_x, xlim), _get_range(_y, ylim)
                ax.set(xlim=xlim, ylim=ylim)
        if text is not None:
            _text.set_text(text[idx])
        return lines

    anim = _animation.FuncAnimation(
        fig, animate, frames=len(profiles), **func_animation_kwargs
    )
    anim.save(filepath, extra_args=['-vcodec', 'libx264'], **save_kwargs)
    plt.close()
    if tqdm is not None:
        pbar.close()

    return anim


def animation_particles(
    *,
    filename: Union[str, Path],
    snaps: List[SnapLike],
    fig: Any = None,
    adaptive_limits: bool = True,
    text: List[str] = None,
    text_kwargs: Dict[str, Any] = {},
    func_animation_kwargs: Dict[str, Any] = {},
    save_kwargs: Dict[str, Any] = {},
    **kwargs,
):
    """Generate an animation of particle plots.

    Parameters
    ----------
    filename
        The file name to save the animation to.
    snaps
        A list of Snap objects to animate.
    fig : optional
        A matplotlib Figure object to animate. If None, generate a new
        Figure.
    adaptive_limits : optional
        If True, adapt plot limits during animation. If False, the plot
        limits are fixed by the initial plot. Default is True.
    text : optional
        List of strings to display per snap.
    text_kwargs : optional
        Keyword arguments to pass to matplotlib text.
    func_animation_kwargs : optional
        Keyword arguments to pass to matplotlib FuncAnimation.
    save_kwargs : optional
        Keyword arguments to pass to matplotlib Animation.save.
    **kwargs
        Arguments to pass to visualize.plot.

    Returns
    -------
    anim
        The matplotlib FuncAnimation object.

    Examples
    --------
    Make an animation of x vs density on the particles.

    >>> plonk.animation(
    ...     filename='animation.mp4',
    ...     snaps=snaps,
    ...     x='x',
    ...     y='density',
    ...     adaptive_limits=False,
    ...     save_kwargs={'fps': 10, 'dpi': 300},
    ... )
    """
    filepath = Path(filename)
    if filepath.suffix != '.mp4':
        raise ValueError('filename should end in ".mp4"')

    if fig is None:
        ax = snaps[0].plot(**kwargs)
        fig = ax.figure
        lines = ax.lines
    else:
        ax = fig.axes[0]
        lines = [line for line in ax.lines]

    if text is not None:
        _text = ax.text(
            0.9, 0.9, text[0], ha='right', transform=ax.transAxes, **text_kwargs
        )

    if tqdm is not None:
        pbar = tqdm(total=len(snaps))
    else:
        logger.info(
            'progress bar not available\ntry pip install tqdm --or-- conda install tqdm'
        )

    x = kwargs.get('x', 'x')
    y = kwargs.get('y', 'y')
    c = kwargs.get('c')
    s = kwargs.get('s')
    units = kwargs.get('units')

    def animate(idx):
        snap = snaps[idx]
        if tqdm is not None:
            pbar.update(n=1)
        xlim, ylim = (0, 0), (0, 0)

        if c is None and s is None:
            try:
                subsnaps = snap.subsnaps_as_list(squeeze=False)
            except AttributeError:
                subsnaps = [snap]
        else:
            subsnaps = [snap]

        for line, subsnap in zip(lines, subsnaps):
            _x, _y, _c, _s, _units = viz._plot_data(
                snap=subsnap, x=x, y=y, c=c, s=s, units=units
            )
            line.set_data(_x, _y)
            if adaptive_limits:
                xlim, ylim = _get_range(_x, xlim), _get_range(_y, ylim)
                line.axes.set(xlim=xlim, ylim=ylim)
        if text is not None:
            _text.set_text(text[idx])
        return lines

    anim = _animation.FuncAnimation(
        fig, animate, frames=len(snaps), **func_animation_kwargs
    )
    anim.save(filepath, extra_args=['-vcodec', 'libx264'], **save_kwargs)
    plt.close()
    if tqdm is not None:
        pbar.close()

    return anim


def _get_range(vals, vlim):
    vmin, vmax = vals.min(), vals.max()
    dv = 0.05 * (vmax - vmin)
    return (
        vmin - dv if vmin - dv < vlim[0] else vlim[0],
        vmax + dv if vmax + dv > vlim[1] else vlim[1],
    )



"""SPH simulation data.

It contains Plonk the implementation of smoothed particle hydrodynamics
combined simulation data. This includes snapshots and time series data.
"""

from .evolution import load_ev
from .simulation import Simulation, load_sim

__all__ = ['Simulation', 'load_ev', 'load_sim']



"""Time series data for global or sink quantites.

This module contains a function for loading global quantities and sink
particle time series data typical of Phantom simulations. These files
track averaged quantities that are more frequently output than snapshot
files.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List, Tuple, Union

from pandas import DataFrame

from ._phantom_evolution import evolution_units as evolution_units_phantom
from ._phantom_evolution import load_data_from_file as load_data_from_file_phantom

if TYPE_CHECKING:
    from .simulation import Simulation


def load_ev(
    filenames: Union[str, Path, Tuple[str], Tuple[Path], List[str], List[Path]],
    data_source: str = 'Phantom',
    config: Union[str, Path] = None,
) -> DataFrame:
    """Load time evolution data from file(s).

    Evolution files track global quantities, such as energy, momentum,
    and density, over time. The time increments in these files is
    smaller than the snapshot file output time. These files are
    typically stored as text files.

    The data is stored as a pandas DataFrame.

    Parameters
    ----------
    filename(s)
        Collection of paths to evolution file(s) in chronological order.
        These should all contain the same columns.
    data_source : optional
        The code used to produce the data. Default is 'Phantom'.
    config : optional
        The path to a Plonk config.toml file.

    Returns
    -------
    dataframe
        A pandas DataFrame with the time evolution data.

    Examples
    --------
    Reading a single evolution file into a pandas DataFrame.

    >>> file_name = 'simulation.ev'
    >>> ev = plonk.load_ev(file_name)

    Reading a collection of evolution files into a pandas DataFrame.

    >>> file_names = ('sim01.ev', 'sim02.ev', 'sim03.ev')
    >>> ev = plonk.load_ev(file_names)
    """
    if data_source == 'Phantom':
        return load_data_from_file_phantom(filenames=filenames, config=config)
    raise ValueError('Cannot determine code used to produce evolution data')


def evolution_units(
    sim: Simulation, data_source: str = 'Phantom', config: Union[str, Path] = None
):
    """Get units of time series data from Simulation object.

    Parameters
    ----------
    sim
        The Simulation object.
    data_source : optional
        The code used to produce the data. Default is 'Phantom'.
    config : optional
        The path to a Plonk config.toml file.

    Returns
    -------
    Dict
    """
    if data_source == 'Phantom':
        return evolution_units_phantom(sim=sim, config=config)
    raise ValueError('Cannot determine code used to produce evolution data')
